# -*- coding: utf-8 -*-

cimport numpy as np
from ultron.sentry.Math.Accumulators.IAccumulators cimport Accumulator


cdef class Diff(Accumulator):

    cdef public double _curr
    cdef public double _previous
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class SimpleReturn(Accumulator):

    cdef public double _diff
    cdef public double _curr
    cdef public double _previous
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class LogReturn(Accumulator):

    cdef public double _diff
    cdef public double _curr
    cdef public double _previous
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class PositivePart(Accumulator):

    cdef public double _pos
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class NegativePart(Accumulator):

    cdef public double _neg
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Max(Accumulator):

    cdef public double _currentMax
    cdef public int _first
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Maximum(Accumulator):

    cdef public double _currentMax
    cdef Accumulator _x
    cdef Accumulator _y

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Min(Accumulator):

    cdef public double _currentMin
    cdef public int _first
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Minimum(Accumulator):

    cdef public double _currentMin
    cdef public int _first
    cdef Accumulator _x
    cdef Accumulator _y

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Sum(Accumulator):

    cdef public double _currentSum
    cdef public int _first
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Average(Accumulator):

    cdef public double _currentSum
    cdef public int _currentCount
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class XAverage(Accumulator):

    cdef public double _average
    cdef public double _exp
    cdef public int _count
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Variance(Accumulator):

    cdef public double _currentSum
    cdef public double _currentSumSquare
    cdef public int _currentCount
    cdef public bint _isPop
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Product(Accumulator):

    cdef public double _product
    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)



# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import pdb
# 固定顺序的施密特正交 (正交 + 归一)
'''
确定因子正交顺序，正交顺序不同，最终得到的因子不同.
'''
def schmidt(factors, class_col):
    class_mkt = factors[class_col]
    factors1 = factors.drop(class_col,axis = 1)
    col_name = factors1.columns
    factors1 = factors1.values
    R = np.zeros((factors1.shape[1], factors1.shape[1]))
    Q = np.zeros(factors1.shape)
    for k in range(0, factors1.shape[1]):
        R[k, k] = np.sqrt(np.dot(factors1[:, k], factors1[:, k]))
        Q[:, k] = factors1[:, k]/R[k, k]
        for j in range(k+1, factors1.shape[1]):
            R[k, j] = np.dot(Q[:, k], factors1[:, j])
            factors1[:, j] = factors1[:, j] - R[k, j]*Q[:, k]

    Q = pd.DataFrame(Q,columns = col_name,index = factors.index)
    Q = pd.concat([Q,class_mkt],axis = 1)
    return Q

# 规范正交
'''
 与主成分分析一样，相对主成分分析，规范正交即可用在截面上，但用在时间序列上就会出现对应关系不一致的问题
 确定因子正交顺序，正交顺序不同，最终得到的因子不同.
'''
def canonial(factors, class_col):
    class_mkt = factors[class_col]
    factors1 = factors.drop(class_col,axis = 1)
    col_name = factors1.columns     
    D,U=np.linalg.eigh(np.dot(factors1.T,factors1))
    D = np.where(D > 0.0000001, D, np.where(D < -0.0000001,D, 0))
    S = np.dot(U,np.diag(D**(-0.5)))
        
    fhat = np.dot(factors1,S)
    fhat = pd.DataFrame(fhat,columns = col_name,index = factors.index)
    fhat = pd.concat([fhat,class_mkt],axis = 1)        

    return fhat

# 对称正交
def symmetry(factors, class_col):
    class_mkt = factors[class_col]
    factors1 = factors.drop(class_col,axis = 1)
    col_name = factors1.columns
    D,U=np.linalg.eigh(np.dot(factors1.T,factors1))
    D = np.where(D > 0.0000001, D, np.where(D < -0.0000001,D, 0.0000001))
    S = np.dot(U,np.diag(D**(-0.5)))
    fhat = np.dot(factors1,S)
    fhat = np.dot(fhat,U.T)
    fhat = pd.DataFrame(fhat,columns = col_name,index = factors.index)
    fhat = pd.concat([fhat,class_mkt],axis = 1)        
    
    return fhat

#
def orthogonalize(factors, class_col):
    class_mkt = factors[class_col]
    factors1 = factors.drop(class_col,axis = 1)
    orth_df = factors1.copy()
    for k in range(0, factors1.shape[1]):
        f = factors1.columns[k]
        f_orth_list = factors1.columns[:k]
        x = np.linalg.lstsq(orth_df[f_orth_list].values, np.matrix(orth_df[f]).T)[0]
        orth_df[f] = (orth_df[f] - orth_df[f_orth_list].dot(x)[0])
    return pd.concat([orth_df,class_mkt],axis = 1)  
    


# -*- coding: utf-8 -*-
cimport cython
from ultron.sentry.Math.Distributions.norm cimport pdf
from ultron.sentry.Math.Distributions.norm cimport cdf
from ultron.sentry.Math.Distributions.norm cimport cdf_derivative
from libc.math cimport exp
from libc.math cimport log
from libc.math cimport sqrt
from ultron.sentry.Math.MathConstants cimport M_SQRT_2
from ultron.sentry.Math.MathConstants cimport M_1_SQRTPI
from ultron.sentry.Math.MathConstants cimport QL_EPSILON
from ultron.sentry.Math.MathConstants cimport M_SQRT2
from ultron.sentry.Math.MathConstants cimport M_SQRTPI


cdef double _M_SQRT_2 = M_SQRT_2
cdef double _M_1_SQRTPI = M_1_SQRTPI
cdef double _QL_EPSILON = QL_EPSILON
cdef double _M_SQRT2 = M_SQRT2
cdef double _M_SQRTPI = M_SQRTPI


cdef class NormalDistribution(object):

    cdef double _average
    cdef double _sigma
    cdef double _normalizationFactor
    cdef double _derNormalizationFactor
    cdef double _denominator

    def __init__(self, average=0.0, sigma=1.0):
        self._average = average
        self._sigma = sigma
        self._normalizationFactor = _M_SQRT_2 * _M_1_SQRTPI / self._sigma
        self._derNormalizationFactor = self._sigma * self._sigma
        self._denominator = 2.0 * self._derNormalizationFactor

    @property
    def average(self):
        return self._average

    @property
    def sigma(self):
        return self._sigma

    def __call__(self, double x):
        x = x - self._average
        return pdf(x, self._denominator, self._normalizationFactor)

    @cython.cdivision(True)
    def derivative(self, double x):
        cdef double deltax = x - self._average
        cdef double exponent = -(deltax * deltax) / self._denominator
        cdef double normVal = 0.0 if exponent <= -690.0 else self._normalizationFactor * exp(exponent)
        return -normVal * deltax / self._derNormalizationFactor


    def __deepcopy__(self, memo):
        return NormalDistribution(self._average, self._sigma)

    def __reduce__(self):
        d = {}

        return NormalDistribution, (self._average, self._sigma), d

    def __setstate__(self, state):
        pass

    def __richcmp__(NormalDistribution self, NormalDistribution other, int op):
        if op == 2:
            return self._average == other.average and self._sigma == other.sigma


cdef class CumulativeNormalDistribution(object):

    cdef double _average
    cdef double _sigma

    def __init__(self, average=0.0, sigma=1.0):
        self._average = average
        self._sigma = sigma

    @property
    def average(self):
        return self._average

    @property
    def sigma(self):
        return self._sigma

    @cython.cdivision(True)
    def __call__(self, double z):
        z = (z - self._average) / self._sigma
        return cdf(z)

    @cython.cdivision(True)
    def derivative(self, z):
        cdef double zn = (z - self._average) / self._sigma
        return cdf_derivative(zn) / self._sigma

    def __deepcopy__(self, memo):
        return CumulativeNormalDistribution(self._average, self._sigma)

    def __reduce__(self):
        d = {}

        return CumulativeNormalDistribution, (self._average, self._sigma), d

    def __setstate__(self, state):
        pass

    def __richcmp__(CumulativeNormalDistribution self, CumulativeNormalDistribution other, int op):
        if op == 2:
            return self._average == other.average and self._sigma == other.sigma


cdef double _a1 = -3.969683028665376e+01
cdef double _a2 = 2.209460984245205e+02
cdef double _a3 = -2.759285104469687e+02
cdef double _a4 = 1.383577518672690e+02
cdef double _a5 = -3.066479806614716e+01
cdef double _a6 = 2.506628277459239e+00

cdef double _b1 = -5.447609879822406e+01
cdef double _b2 = 1.615858368580409e+02
cdef double _b3 = -1.556989798598866e+02
cdef double _b4 = 6.680131188771972e+01
cdef double _b5 = -1.328068155288572e+01

cdef double _c1 = -7.784894002430293e-03
cdef double _c2 = -3.223964580411365e-01
cdef double _c3 = -2.400758277161838e+00
cdef double _c4 = -2.549732539343734e+00
cdef double _c5 = 4.374664141464968e+00
cdef double _c6 = 2.938163982698783e+00

cdef double _d1 = 7.784695709041462e-03
cdef double _d2 = 3.224671290700398e-01
cdef double _d3 = 2.445134137142996e+00
cdef double _d4 = 3.754408661907416e+00

# Limits of the approximation regions
cdef double _x_low = 0.02425
cdef double _x_high = 1.0 - _x_low


@cython.cdivision(True)
cdef double _tail_value(double x):
    cdef double z

    if x < _x_low:
        # Rational approximation for the lower region 0<x<u_low
        z = sqrt(-2.0 * log(x))
        z = (((((_c1 * z + _c2) * z + _c3) * z + _c4) * z + _c5) * z + _c6) \
            / ((((_d1 * z + _d2) * z + _d3) * z + _d4) * z + 1.0)
    else:
        # Rational approximation for the upper region u_high<x<1
        z = sqrt(-2.0 * log(1.0 - x))
        z = -(((((_c1 * z + _c2) * z + _c3) * z + _c4) * z + _c5) * z + _c6) \
            / ((((_d1 * z + _d2) * z + _d3) * z + _d4) * z + 1.0)
    return z


cdef class InverseCumulativeNormal(object):

    def __init__(self, average=0.0, sigma=1.0, fullAccuracy=False):
        self._average = average
        self._sigma = sigma
        self._fullAcc = fullAccuracy

    @property
    def average(self):
        return self._average

    @property
    def sigma(self):
        return self._sigma

    @property
    def isFullAcc(self):
        return self._fullAcc

    @cython.cdivision(True)
    cdef double _standard_value(self, double x):

        cdef double z
        cdef double r

        if x < _x_low or x > _x_high:
            z = _tail_value(x)
        else:
            z = x - 0.5
            r = z * z
            z = (((((_a1 * r + _a2) * r + _a3) * r + _a4) * r + _a5) * r + _a6) * z \
                / (((((_b1 * r + _b2) * r + _b3) * r + _b4) * r + _b5) * r + 1.0)
        if self._fullAcc:
            r = (cdf(z) - x) * _M_SQRT2 * _M_SQRTPI * exp(0.5 * z * z)
            z -= r / (1 + 0.5 * z * r)
        return z

    cdef double inv(self, double x):
        return self._average + self._sigma * self._standard_value(x)

    def __call__(self, x):
        return self.inv(x)

    def __deepcopy__(self, memo):
        return InverseCumulativeNormal(self._average, self._sigma, self._fullAcc)

    def __reduce__(self):
        d = {}

        return InverseCumulativeNormal, (self._average, self._sigma, self._fullAcc), d

    def __setstate__(self, state):
        pass

    def __richcmp__(InverseCumulativeNormal self, InverseCumulativeNormal other, int op):
        if op == 2:
            return self._average == other.average and self._sigma == other.sigma and self._fullAcc == self.isFullAcc



from typing import Iterable
from typing import Dict
import pandas as pd
import numpy as np
import pickle, pdb, time, itertools, datetime,hashlib
from copy import copy
from joblib import Parallel, delayed
from ultron.factor.genetic.geneticist.operators import expression_params,calc_factor
from ultron.factor.fitness.weighted import Weighted
from ultron.factor.combine.combine_engine import CombineEngine
from ultron.utilities.utils import check_random_state
from ultron.utilities.jobs import partition_estimators
from ultron.utilities.utils import check_random_state

import warnings
warnings.filterwarnings("ignore")

MAX_INT = np.iinfo(np.int32).max
MIN_INT = np.iinfo(np.int32).min 

class CombineAttr(object):
    def __init__(self, name, method, span, custom_params):
        self._name = name
        self._method = method
        self._span = span
        self._custom_params = custom_params
         
    def log(self):
        print(self._name, self._method, self._span, self._custom_params)
            
class Individual(object):
    def __init__(self, value, stype):
        self._value = value
        self._stype = stype #   1. 函数  2.因子(可用于交叉) 
        
class Program(object):
    def __init__(self, gen, random_state, init_depth, p_point_replace,
                 factor_sets, span_sets, half_sets, weight_limit_sets,
                program=None, parents=None):
        self._gen = gen
        self._random_state = random_state
        self._init_depth = init_depth
        self._p_point_replace = p_point_replace
        self._program = program
        self._parents = parents
        self._raw_fitness = 0
        self._is_valid = True
        self._create_time = datetime.datetime.now()
        self._factor_sets = factor_sets
        self.init_combine_func(span_sets, half_sets, weight_limit_sets)
        self._risk_styles = ['BETA','MOMENTUM','SIZE','EARNYILD','RESVOL','GROWTH','BTOP','LEVERAGE','LIQUIDTY','SIZENL']
        self._industry_styles = ['Bank','RealEstate','Health','Transportation','Mining',
                                 'NonFerMetal','HouseApp','LeiService','MachiEquip','BuildDeco',
                                 'CommeTrade','CONMAT','Auto','Textile','FoodBever','Electronics',
                                 'Computer','LightIndus','Utilities','Telecom','AgriForest','CHEM',
                                 'Media','IronSteel','NonBankFinan','ELECEQP','AERODEF','Conglomerates']
        self._name = 'ultron_' + str(int(time.time() * 1000000 + datetime.datetime.now().microsecond))
        if self._program is None:
            self._program = self.build_program(random_state)
        self.create_identification()
        
    def init_combine_func(self, span_sets, half_sets, weight_limit_sets):
        self._combine_func = []
        hist_func_sets = ['hist_ret_combine','hist_ic_combine']
        max_func_sets = ['max_icir_combine','max_ic_combine']
        hist_method_sets = ['equal','half_life']
        max_method_sets = ['sample','shrunk']
        for name in hist_func_sets:
            for method in hist_method_sets:
                for span in span_sets:
                    for half in half_sets:
                        self._combine_func.append(CombineAttr(name, method, span, half))
        for name in max_func_sets:
            for method in max_method_sets:
                for span in span_sets:
                    for weight in weight_limit_sets:
                        self._combine_func.append(CombineAttr(name, method, span, weight))
    
    def log(self):
        parents = {'method':'gen'} if self._parents is None else self._parents
        name = self._name
        sets = [p._value for p in self._program if p._stype == 2]
        func = [p._value for p in self._program if p._stype == 1][0]
        sets += [func._name, func._method, str(func._span), str(func._custom_params)]
        print('name:%s,raw_fitness:%f, program:%s,create_time:%s,identification:%s' % (
            self._name, self._raw_fitness, sets, self._create_time,self._identification))
    
    def _func_method(self):
        func_sets = [p for p in self._program if p._stype == 1]
        return func_sets[0]._value
        
    def _factors_sets(self):
        return [p._value for p in self._program if p._stype == 2]
    
    def create_identification(self):
        m = hashlib.md5()
        program = [p._value for p in self._program if p._stype == 2]
        func = [p._value for p in self._program if p._stype == 1][0]
        program += [func._name, func._method, str(func._span), str(func._custom_params)]
        program.sort()
        m.update(bytes(''.join(program), encoding='UTF-8'))
        self._identification = m.hexdigest()
        
    def duplicate_removal(self, program):
        #重复的函数剔除，若被正交
        factors_sets = [p._value for p in program if p._stype == 2]
        factors_sets = list(set(factors_sets))
        func_sets = [p for p in program if p._stype == 1]
        new_program = []
        if len(func_sets) == 0:
            combine_func = self._combine_func[np.random.randint(len(self._combine_func))]
            new_program = [Individual(combine_func,1)]
            new_program += [Individual(factor, 2) for factor in factors_sets]
        elif len(func_sets) > 0:
            retain_func = func_sets[np.random.randint(len(func_sets))]
            new_program = [retain_func]
            new_program += [Individual(factor, 2) for factor in factors_sets]
        return new_program
    
    def build_program(self, random_state):
        factors_max = self._init_depth[1] if self._init_depth[1] < len(self._factor_sets) else len(self._factor_sets)
        factors_max_prob = np.random.randint(factors_max) / len(self._factor_sets)
        factors = self._factor_sets[np.random.randint(len(self._factor_sets))]
        combine_func = self._combine_func[np.random.randint(len(self._combine_func))]
        program = [Individual(factors, 2), Individual(combine_func, 1)]
        factors_list = (np.random.uniform(0, 1, (1, len(self._factor_sets))) < factors_max_prob).tolist()
        factors_list = list(set(list(np.array(self._factor_sets)[factors_list])))
        program += [Individual(factor, 2) for factor in factors_list]
        return self.duplicate_removal(program)
    
    
    def get_subtree(self, random_state=None, program=None):
        if program is None:
            program = self._program
        probs = np.array([0.9 if node._value in self._factor_sets else 0.1 for node in program])
        probs = np.cumsum(probs / probs.sum())
        start = np.searchsorted(probs, np.random.uniform())
        end = start if (len(program) - 1) <= start else np.random.randint(start, len(program) - 1)
        return start, end
    
    def reproduce(self):
        return copy(self._program)
    
    def crossover(self, donor, random_state):
        start, end = self.get_subtree(random_state)
        end -= 1
        removed = range(start, end)
        donor_start, donor_end = self.get_subtree(random_state, donor)
        donor_removed = list(set(range(len(donor))) -
                             set(range(donor_start, donor_end)))
        return self.duplicate_removal(self._program[:start] +
                donor[donor_start:donor_end] +
                self._program[end:]), removed, donor_removed
    
    ##树变异        
    def subtree_mutation(self, random_state=None):
        chicken = self.build_program(random_state)
        return self.crossover(chicken, random_state)
    
    ##突变异
    def hoist_mutation(self, random_state):
        start, end = self.get_subtree(random_state)
        subtree = self._program[start:end]
        sub_start, sub_end = self.get_subtree(random_state, subtree)
        hoist = subtree[sub_start:sub_end]
        removed = list(set(range(start, end)) -
                       set(range(start + sub_start, start + sub_end)))
        return self.duplicate_removal(self._program[:start] + hoist + self._program[end:]),removed
    
    def point_mutation(self, random_state):
        program = copy(self._program)
        mutate = np.where(np.random.uniform(size=len(program)) <
                          self._p_point_replace)[0]
        for node in mutate:
            if program[node]._value in self._factor_sets:
                replace_node = self._factor_sets[np.random.randint(0,len(self._factor_sets)-1)]
                program[node] = Individual(replace_node,2)
            else:
                replace_node = self._combine_func[np.random.randint(0,len(self._combine_func)-1)]
                program[node] = Individual(replace_node,1)
        return self.duplicate_removal(program), list(mutate)
    
    def evaluation(self, standard_data,  default_value, fitness_method='high_frequency',
                   horizon=2, is_neutralize=0):
        risk_data = standard_data[['trade_date','code'] + self._industry_styles + self._risk_styles]
        mkt_df = standard_data[['closePrice', 'openPrice', 'lowestPrice', 'highestPrice', 'turnoverVol',
                         'turnoverValue', 'trade_date', 'code', 'accumAdjFactor',
                         'chgPct', 'bar30_vwap', 'bar60_vwap']]
        #合成
        func_sets = self._func_method()
        factors_columns = self._factors_sets()
        combine = CombineEngine.create_engine(func_sets._name)
        try:
            if func_sets._name == 'hist_ret_combine' or func_sets._name == 'hist_ic_combine':
                if func_sets._method == 'equal':
                    factors_data, hist_ret = combine(factor_df=standard_data, mret_df=mkt_df, factor_list=factors_columns,
                                                     span=func_sets._span)
                elif func_sets._method == 'half_life':
                    factors_data, hist_ret = combine(factor_df=standard_data, mret_df=mkt_df, factor_list=factors_columns,
                                                         span=func_sets._span, method='half_life', 
                                                     half_life=func_sets._custom_params)
            elif func_sets._name == 'max_icir_combine' or func_sets._name == 'max_ic_combine':
                if func_sets._method == 'sample':
                    factors_data, m_ir_df = combine(factor_df=standard_data, mret_df=mkt_df, factor_list=factors_columns, 
                                                    span=func_sets._span)
                elif func_sets._method == 'shrunk':
                    factors_data, m_ir_df = combine(factor_df=standard_data, mret_df=mkt_df, factor_list=factors_columns,
                                                    span=func_sets._span, method='shrunk', weight_limit=func_sets._custom_params)
                
                
            weight = Weighted.create_weighted(method=fitness_method)
            result =  weight.run(factor_data=factors_data, horizon=horizon,
                               risk_data=risk_data, mkt_df=mkt_df,
                               factor_name='combine', is_neutralize=is_neutralize,
                               default_value=np.iinfo(np.int32))
            raw_fitness = default_value if result['status'] == 0 else abs(result['score'])
            self._raw_fitness = default_value if np.isnan(raw_fitness) else raw_fitness
            self._is_valid = False if self._raw_fitness == default_value else True
        except Exception as e:
            self._raw_fitness = default_value
            self._is_valid  = False
            print(e)

def parallel_evolve(n_programs, parents, standard_data, seeds, factor_sets, span_sets, 
                    half_sets, weight_limit_sets, gen, params):
    tournament_size = params['tournament_size']
    init_depth = params['init_depth']
    greater_is_better = params['greater_is_better'] 
    method_probs = params['method_probs']
    p_point_replace = params['p_point_replace']
    
    programs = []
    def _contenders(tour_parents):
        ctournament_size = tournament_size if tournament_size < 2 else 2
        contenders = random_state.randint(0, len(tour_parents), ctournament_size)
        return [tour_parents[p] for p in contenders]
    
    def _tournament(tour_parents):
        ttournament_size = tournament_size if tournament_size < len(tour_parents) else len(tour_parents)
        contenders = random_state.randint(0, len(tour_parents), ttournament_size)
        raw_fitness = [tour_parents[p]._raw_fitness for p in contenders]
        if greater_is_better:
            parent_index = contenders[np.argmax(raw_fitness)]
        else:
            parent_index = contenders[np.argmin(raw_fitness)]
        return tour_parents[parent_index], parent_index
    
    for i in range(n_programs):
        random_state = check_random_state(seeds[i])
        if parents is None:
            program = None
            genome = None
        else:
            method = random_state.uniform()
            parent, parent_index = _tournament(copy(parents))
            ori_parent = copy(parent)
            contenders = _contenders(copy(parents))
            for contender in contenders:
                program, removed, remains = parent.crossover(contender._program, random_state)
                parent = Program(gen=gen, random_state=random_state, factor_sets=factor_sets, 
                                 span_sets=span_sets, half_sets=half_sets,weight_limit_sets=weight_limit_sets,
                                 init_depth=init_depth,p_point_replace=p_point_replace, 
                                 program=program, parents=parents)
                
            #新特征种群加入
            if random_state.uniform() < method_probs[2]:
                program = Program(gen=gen, random_state=random_state, factor_sets=factor_sets, 
                                  span_sets=span_sets, half_sets=half_sets,weight_limit_sets=weight_limit_sets,
                                  init_depth=init_depth,p_point_replace=p_point_replace, 
                                  program=None)
                program, removed, remains = parent.crossover(program._program, random_state)
                parent = Program(gen=gen, random_state=random_state, factor_sets=factor_sets, 
                                 span_sets=span_sets, half_sets=half_sets,weight_limit_sets=weight_limit_sets,
                                 init_depth=init_depth,p_point_replace=p_point_replace, 
                                 program=program)
      
            if method < 0:#method_probs[0]: # # crossover
                donor, donor_index = _tournament(copy(parents))
                program, removed, remains = parent.crossover(donor._program, random_state)
                genome = {'method':'Crossover',
                         'parent_idx':parent_index,
                         'parent_nodes':removed,
                         'donor_idx':donor_index,
                         'donor_nodes':remains}
            elif method < 0:#method_probs[1]: # subtree_mutation
                program, removed, _ = parent.subtree_mutation(random_state)
                genome = {'method': 'Subtree Mutation',
                          'parent_idx': parent_index,
                          'parent_nodes': removed}
            elif method < 0:#method_probs[2]: # hoist_mutation
                program, removed = parent.hoist_mutation(random_state)
                genome = {'method': 'Hoist Mutation',
                          'parent_idx': parent_index,
                          'parent_nodes': removed}
            elif method < 1:#method_probs[3]: # point_mutation
                program,mutated = parent.point_mutation(random_state)
                genome = {'method': 'Point Mutation',
                          'parent_idx': parent_index,
                          'parent_nodes': mutated}
            else:
                program = parent.reproduce() # reproduction
                genome = {'method': 'Reproduction',
                          'parent_idx': parent_index,
                          'parent_nodes': []}
                
           
            # 与原始自身进行交叉
            if random_state.uniform() < method_probs[3]:
                program = Program(gen=gen, random_state=random_state, factor_sets=factor_sets, 
                                  span_sets=span_sets, half_sets=half_sets,weight_limit_sets=weight_limit_sets,
                                  init_depth=init_depth,p_point_replace=p_point_replace, 
                                  program=program, parents=genome)
                program, removed, remains = program.crossover(ori_parent._program, random_state)
                
        program = Program(gen=gen, random_state=random_state, factor_sets=factor_sets, 
                          span_sets=span_sets, half_sets=half_sets,weight_limit_sets=weight_limit_sets,
                          init_depth=init_depth,p_point_replace=p_point_replace, 
                          program=program, parents=genome)
        default_value = MIN_INT if greater_is_better else MAX_INT
        program.evaluation(standard_data,  default_value)
        print(program._raw_fitness)
        programs.append(program)
    return programs
    
class Gentic(object):
    def __init__(self, factors_sets, span_sets, half_sets, weight_limit_sets=[True, False],
                 random_state=None, population_size=5, tournament_size=2, 
                 generations=MAX_INT,greater_is_better=True,verbose=1,convergence=0.05,
                 init_depth=(5, 6), n_jobs=1, p_crossover=0.9, p_subtree_mutation=0.01,
                 p_hoist_mutation=0.01, p_point_mutation=0.01, p_point_replace=0.5):
        self._random_state = random_state
        self._generations = MAX_INT if generations == 0 else generations
        self._population_size = population_size
        self._tournament_size = tournament_size
        self._init_depth = init_depth
        self._n_jobs = n_jobs
        self._greater_is_better = greater_is_better
        self._p_crossover = p_crossover
        self._p_subtree_mutation = p_subtree_mutation
        self._p_hoist_mutation = p_hoist_mutation
        self._p_point_mutation = p_point_mutation
        self._p_point_replace = p_point_replace
        self._convergence = convergence
        self._verbose = verbose
        self._factors_sets = factors_sets
        self._span_sets = span_sets
        self._half_sets = half_sets
        self._weight_limit_sets = weight_limit_sets
        
    def train(self, standard_data):
        random_state = check_random_state(self._random_state)
        self._method_probs = np.array([self._p_crossover,
                                       self._p_subtree_mutation,
                                       self._p_hoist_mutation,
                                       self._p_point_mutation])
        
        self._method_probs = np.cumsum(self._method_probs)
        
        if self._method_probs[-1] > 1:
            raise ValueError('The sum of p_crossover, p_subtree_mutation, '
                             'p_hoist_mutation and p_point_mutation should '
                             'total to 1.0 or less.')
            
        if (isinstance(self._init_depth, tuple) and len(self._init_depth) != 2):
            raise ValueError('init_depth should be a tuple with length two.')
        
        if (isinstance(self._init_depth, tuple) and (self._init_depth[0] > self._init_depth[1])):
            raise ValueError('init_depth should be in increasing numerical '
                             'order: (min_depth, max_depth).')
            
        params = {}
        params['tournament_size'] = self._tournament_size
        params['p_point_replace'] = self._p_point_replace
        params['init_depth'] = self._init_depth
        params['greater_is_better'] = self._greater_is_better
        params['method_probs'] = self._method_probs
        params['p_point_replace'] = self._p_point_replace
        
        count_time = 0
        iteration_best_programs = None
        self._programs = []
        run_details = {'generation': [],
                             'average_fitness': [],
                             'best_fitness': [],
                             'generation_time': [],
                             'best_programs':[]}
        prior_generations = len(self._programs)
        n_more_generations = self._generations - prior_generations
        for gen in range(prior_generations, self._generations):
            start_time = time.time()
            if gen == 0:
                parents = None
            else:
                parents = self._programs[gen - 1]
                
            n_jobs, n_programs, starts = partition_estimators(
                self._population_size, self._n_jobs)
            
            seeds = random_state.randint(MAX_INT, size=self._population_size)
            population = Parallel(n_jobs=self._n_jobs,
                                  verbose=self._verbose)(
                delayed(parallel_evolve)(n_programs[i], parents, standard_data, seeds, self._factors_sets,
                                         self._span_sets, self._half_sets, self._weight_limit_sets, 
                                         gen, params)
                for i in range(n_jobs))
            
            population = list(itertools.chain.from_iterable(population))
            population = [p for p in population if p._is_valid]
            self._programs.append(population)
            
            if iteration_best_programs is not None:
                programs = np.concatenate([population,iteration_best_programs])
            else:
                programs = population
                
             
            fitness = [program._raw_fitness for program in programs]
            if self._greater_is_better:
                iteration_best_programs = np.array(programs)[np.argsort(fitness)[-self._tournament_size:]]
            else:
                iteration_best_programs = np.array(programs)[np.argsort(fitness)[:self._tournament_size:]]
            
            fitness = [program._raw_fitness for program in iteration_best_programs]
            
            for program in iteration_best_programs:
                program.log()
                
            run_details['generation'].append(gen)
            run_details['average_fitness'].append(np.mean(fitness))
            generation_time = time.time() - start_time
            run_details['generation_time'].append(generation_time)
            run_details['best_programs'].append(iteration_best_programs)
            
            print(np.mean(fitness))
            
            if gen ==0:
                continue
             
            d_value = np.mean(fitness) - run_details['average_fitness'][gen - 1]
            print('d_value:%f,convergence:%f,count:%d,con_time:%d' % (d_value, self._convergence,
                                                                      len(iteration_best_programs), count_time))
            if abs(d_value) < self._convergence:
                count_time += 1
                if count_time > 3:
                    break
            else:
                count_time = 0
            
        return run_details


# -*- coding: utf-8 -*-

cdef class Date(object):

    cdef public int __serialNumber__
    cdef public int _year
    cdef public int _month
    cdef public int _day

    cpdef int dayOfMonth(self)

    cpdef int dayOfYear(self)

    cpdef int year(self)

    cpdef int month(self)

    cpdef int weekday(self)

    cpdef toDateTime(self)

    cdef _calculate_date(self, int year, int month, int day)


cpdef check_date(date)


# -*- coding: utf-8 -*-

import numpy as np
import numba as nb
from . utilities import group_mapping
from . utilities import aggregate
from . utilities import transform
from . utilities import array_index
from . utilities import simple_mean
from . utilities import simple_std


@nb.njit(nogil=True, cache=True)
def mask_values_2d(x: np.ndarray,
                   mean_values: np.ndarray,
                   std_values: np.ndarray,
                   num_stds: int = 3) -> np.ndarray:
    res = x.copy()
    length, width = x.shape

    for i in range(length):
        for j in range(width):
            ubound = mean_values[i, j] + num_stds * std_values[i, j]
            lbound = mean_values[i, j] - num_stds * std_values[i, j]
            if x[i, j] > ubound:
                res[i, j] = ubound
            elif x[i, j] < lbound:
                res[i, j] = lbound

    return res


@nb.njit(nogil=True, cache=True)
def mask_values_1d(x: np.ndarray,
                   mean_values: np.ndarray,
                   std_values: np.ndarray,
                   num_stds: int = 3) -> np.ndarray:
    res = x.copy()
    length, width = x.shape

    for j in range(width):
        ubound = mean_values[j] + num_stds * std_values[j]
        lbound = mean_values[j] - num_stds * std_values[j]
        for i in range(length):
            if x[i, j] > ubound:
                res[i, j] = ubound
            elif x[i, j] < lbound:
                res[i, j] = lbound
    return res


def winsorize_normal(x: np.ndarray, num_stds: int = 3, ddof=1, groups: np.ndarray = None) -> np.ndarray:
    if groups is not None:
        groups = group_mapping(groups)
        mean_values = transform(groups, x, 'mean')
        std_values = transform(groups, x, 'std', ddof)
        res = mask_values_2d(x, mean_values, std_values, num_stds)
    else:
        std_values = simple_std(x, axis=0, ddof=ddof)
        mean_values = simple_mean(x, axis=0)
        res = mask_values_1d(x, mean_values, std_values, num_stds)
    return res


class NormalWinsorizer(object):

    def __init__(self, num_stds: int=3, ddof=1):
        self.num_stds = num_stds
        self.ddof = ddof
        self.mean = None
        self.std = None
        self.labels = None

    def fit(self, x: np.ndarray, groups: np.ndarray=None):
        if groups is not None:
            group_index = group_mapping(groups)
            self.mean = aggregate(group_index, x, 'mean')
            self.std = aggregate(group_index, x, 'std', self.ddof)
            self.labels = np.unique(groups)
        else:
            self.mean = simple_mean(x, axis=0)
            self.std = simple_std(x, axis=0, ddof=self.ddof)

    def transform(self, x: np.ndarray, groups: np.ndarray=None) -> np.ndarray:
        if groups is not None:
            index = array_index(self.labels, groups)
            return mask_values_2d(x, self.mean[index], self.std[index], self.num_stds)
        else:
            return mask_values_1d(x, self.mean, self.std, self.num_stds)

    def __call__(self, x: np.ndarray, groups: np.ndarray=None) -> np.ndarray:
        return winsorize_normal(x, self.num_stds, self.ddof, groups)



# -*- coding: utf-8 -*-

from ultron.tradingday.Enums._BizDayConventions cimport BizDayConventions
from ultron.tradingday.Enums._DateGeneration cimport DateGeneration
from ultron.tradingday.Enums._TimeUnits cimport TimeUnits
from ultron.tradingday.Date cimport Date
from ultron.tradingday.Period cimport Period
from ultron.tradingday.Calendar cimport Calendar
from ultron.tradingday.Env import Settings

cdef class Schedule(object):

    def __init__(self,
                 Date effectiveDate,
                 Date terminationDate,
                 Period tenor,
                 Calendar calendar,
                 int convention=BizDayConventions.Following,
                 int terminationConvention=BizDayConventions.Following,
                 int dateGenerationRule=DateGeneration.Forward,
                 bint endOfMonth=False,
                 Date firstDate=None,
                 Date nextToLastDate=None):

        cdef int i
        cdef size_t dateLen
        cdef Calendar nullCalendar
        cdef Date evalDate
        cdef int y
        cdef int periods
        cdef Date seed
        cdef Date temp
        cdef Date exitDate

        # Initialize private data
        self._effectiveDate = effectiveDate
        self._terminationDate = terminationDate
        self._tenor = tenor
        self._cal = calendar
        self._convention = convention
        self._terminationConvention = terminationConvention
        self._rule = dateGenerationRule
        self._dates = []
        self._isRegular = []

        if tenor < Period("1M"):
            self._endOfMonth = False
        else:
            self._endOfMonth = endOfMonth

        if not firstDate or firstDate == effectiveDate:
            self._firstDate = None
        else:
            self._firstDate = firstDate

        if not nextToLastDate or nextToLastDate == terminationDate:
            self._nextToLastDate = None
        else:
            self._nextToLastDate = nextToLastDate

        # in many cases (e.g. non-expired bonds) the effective date is not
        # really necessary. In these cases a decent placeholder is enough
        if not effectiveDate and not firstDate and dateGenerationRule == DateGeneration.Backward:
            evalDate = Settings.evaluationDate
            #require(evalDate < terminationDate, ValueError, "null effective date")
            if nextToLastDate:
                y = int((nextToLastDate - evalDate) / 366) + 1
                effectiveDate = nextToLastDate - Period(length=y, units=TimeUnits.Years)
            else:
                y = int((terminationDate - evalDate) / 366) + 1
                effectiveDate = terminationDate - Period(length=y, units=TimeUnits.Years)
        else:
            pass
            #require(effectiveDate, ValueError, "null effective date")

        #require(effectiveDate < terminationDate, ValueError, "effective date ({0}) "
        #                                                         "later than or equal to termination date ({1}"
        #.format(effectiveDate, terminationDate))

        if tenor.length() == 0:
            self._rule = DateGeneration.Zero
        else:
            pass
            #require(tenor.length() > 0, ValueError, "non positive tenor ({0:d}) not allowed".format(tenor.length()))

        if self._firstDate:
            if self._rule == DateGeneration.Backward or self._rule == DateGeneration.Forward:
                pass
                #require(effectiveDate < self._firstDate < terminationDate, ValueError,
                #            "first date ({0}) out of effective-termination date range [{1}, {2})"
                #            .format(self._firstDate, effectiveDate, terminationDate))
                # we should ensure that the above condition is still
                # verified after adjustment
            elif self._rule == DateGeneration.Zero:
                raise ValueError("first date incompatible with {0:d} date generation rule".format(self._rule))
            else:
                raise ValueError("unknown rule ({0:d})".format(self._rule))

        if self._nextToLastDate:
            if self._rule == DateGeneration.Backward or self._rule == DateGeneration.Forward:
                pass
                #require(effectiveDate < self._nextToLastDate < terminationDate, ValueError,
                #            "next to last date ({0}) out of effective-termination date range [{1}, {2})"
                #            .format(self._nextToLastDate, effectiveDate, terminationDate))
                # we should ensure that the above condition is still
                # verified after adjustment
            elif self._rule == DateGeneration.Zero:
                raise ValueError("next to last date incompatible with {0:d} date generation rule".format(self._rule))
            else:
                raise ValueError("unknown rule ({0:d})".format(self._rule))

        # calendar needed for endOfMonth adjustment
        nullCalendar = Calendar("Null")
        periods = 1

        if self._rule == DateGeneration.Zero:
            self._tenor = Period(length=0, units=TimeUnits.Years)
            self._dates.extend([effectiveDate, terminationDate])
            self._isRegular.append(True)
        elif self._rule == DateGeneration.Backward:
            self._dates.append(terminationDate)
            seed = terminationDate
            if self._nextToLastDate:
                self._dates.insert(0, self._nextToLastDate)
                temp = nullCalendar.advanceDate(seed,
                                                Period(length=-periods * self._tenor.length(), units=self._tenor.units()),
                                                convention, self._endOfMonth)
                if temp != self._nextToLastDate:
                    self._isRegular.insert(0, False)
                else:
                    self._isRegular.insert(0, True)
                seed = self._nextToLastDate

            exitDate = effectiveDate
            if self._firstDate:
                exitDate = self._firstDate

            while True:
                temp = nullCalendar.advanceDate(seed,
                                                Period(length=-periods * self._tenor.length(), units=self._tenor.units()),
                                                convention, self._endOfMonth)
                if temp < exitDate:
                    if self._firstDate and self._cal.adjustDate(self._dates[0], convention) != self._cal.adjustDate(
                            self._firstDate, convention):
                        self._dates.insert(0, self._firstDate)
                        self._isRegular.insert(0, False)
                    break
                else:
                    # skip dates that would result in duplicates
                    # after adjustment
                    if self._cal.adjustDate(self._dates[0], convention) != self._cal.adjustDate(temp, convention):
                        self._dates.insert(0, temp)
                        self._isRegular.insert(0, True)
                    periods += 1

            if self._cal.adjustDate(self._dates[0], convention) != self._cal.adjustDate(effectiveDate, convention):
                self._dates.insert(0, effectiveDate)
                self._isRegular.insert(0, False)

        elif self._rule == DateGeneration.Forward:
            self._dates.append(effectiveDate)

            seed = self._dates[-1]

            if self._firstDate:
                self._dates.append(self._firstDate)
                temp = nullCalendar.advanceDate(seed,
                                                Period(length=periods * self._tenor.length(), units=self._tenor.units()),
                                                convention, self._endOfMonth)
                if temp != self._firstDate:
                    self._isRegular.append(False)
                else:
                    self._isRegular.append(True)
                seed = self._firstDate

            exitDate = terminationDate
            if self._nextToLastDate:
                exitDate = self._nextToLastDate

            while True:
                temp = nullCalendar.advanceDate(seed,
                                                Period(length=periods * self._tenor.length(), units=self._tenor.units()),
                                                convention, self._endOfMonth)
                if temp > exitDate:
                    if self._nextToLastDate and self._cal.adjustDate(self._dates[-1],
                                                                     convention) != self._cal.adjustDate(
                            self._nextToLastDate, convention):
                        self._dates.append(self._nextToLastDate)
                        self._isRegular.append(False)
                    break
                else:
                    # skip dates that would result in duplicates
                    # after adjustment
                    if self._cal.adjustDate(self._dates[-1], convention) != self._cal.adjustDate(temp, convention):
                        self._dates.append(temp)
                        self._isRegular.append(True)
                    periods += 1

            if self._cal.adjustDate(self._dates[-1], terminationConvention) != self._cal.adjustDate(terminationDate,
                                                                                                    terminationConvention):
                self._dates.append(terminationDate)
                self._isRegular.append(False)
        else:
            raise ValueError("unknown rule ({0:d})".format(self._rule))

        # adjustments
        if self._endOfMonth and self._cal.isEndOfMonth(seed):
            # adjust to end of month
            if convention == BizDayConventions.Unadjusted:
                for i in range(len(self._dates) - 1):
                    self._dates[i] = Date.endOfMonth(self._dates[i])
            else:
                for i in range(len(self._dates) - 1):
                    self._dates[i] = self._cal.endOfMonth(self._dates[i])

            if terminationConvention != BizDayConventions.Unadjusted:
                self._dates[0] = self._cal.endOfMonth(self._dates[0])
                self._dates[-1] = self._cal.endOfMonth(self._dates[-1])
            else:
                if self._rule == DateGeneration.Backward:
                    self._dates[-1] = Date.endOfMonth(self._dates[-1])
                else:
                    self._dates[0] = Date.endOfMonth(self._dates[0])
        else:
            for i in range(len(self._dates) - 1):
                self._dates[i] = self._cal.adjustDate(self._dates[i], convention)

            if terminationConvention != BizDayConventions.Unadjusted:
                self._dates[-1] = self._cal.adjustDate(self._dates[-1], terminationConvention)

        # Final safety checks to remove extra next-to-last date, if
        # necessary.  It can happen to be equal or later than the end
        # date due to EOM adjustments (see the Schedule test suite
        # for an example).

        dateLen = len(self._dates)

        if dateLen >= 2 and self._dates[dateLen - 2] >= self._dates[-1]:
            self._isRegular[dateLen - 2] = (self._dates[dateLen - 2] == self._dates[-1])
            self._dates[dateLen - 2] = self._dates[-1]
            self._dates.pop()
            self._isRegular.pop()

        if len(self._dates) >= 2 and self._dates[1] <= self._dates[0]:
            self._isRegular[1] = (self._dates[1] == self._dates[0])
            self._dates[1] = self._dates[0]
            self._dates = self._dates[1:]
            self._isRegular = self._isRegular[1:]

        #require(len(self._dates) >= 1, ValueError, "degenerate single date ({0}) schedule\n"
        #                                               "seed date: {1}\n"
        #                                               "exit date: {2}\n"
        #                                               "effective date: {3}\n"
        #                                               "first date: {4}\n"
        #                                               "next to last date: {5}\n"
        #                                               "termination date: {6}\n"
        #                                               "generation rule: {7}\n"
        #                                               "end of month: {8}\n"
        #.format(self._dates[0],
        #        seed, exitDate,
        #        effectiveDate,
        #        firstDate,
        #        nextToLastDate,
        #        terminationDate,
        #        self._rule, self._endOfMonth))

    cpdef size_t size(self):
        return len(self._dates)

    cpdef Calendar calendar(self):
        return self._cal

    cpdef Period tenor(self):
        return self._tenor

    cpdef bint endOfMonth(self):
        return self._endOfMonth

    cpdef bint isRegular(self, size_t i):
        return self._isRegular[i-1]

    def __getitem__(self, item):
        return self._dates[item]

    def __deepcopy__(self, memo):
        return Schedule(self._effectiveDate,
                        self._terminationDate,
                        self._tenor,
                        self._cal,
                        self._convention,
                        self._terminationConvention,
                        self._rule,
                        self._endOfMonth,
                        self._firstDate,
                        self._nextToLastDate)

    def __reduce__(self):
        d = {}

        return Schedule, (self._effectiveDate,
                          self._terminationDate,
                          self._tenor,
                          self._cal,
                          self._convention,
                          self._terminationConvention,
                          self._rule,
                          self._endOfMonth,
                          self._firstDate,
                          self._nextToLastDate), d

    def __setstate__(self, state):
        pass

    def __richcmp__(self, Schedule other, int op):
        if op == 2:
            return self._effectiveDate == other._effectiveDate \
                   and self._terminationDate == other._terminationDate \
                   and self._tenor == other._tenor \
                   and self._cal == other._cal \
                   and self._convention == other._convention \
                   and self._terminationConvention == other._terminationConvention \
                   and self._rule == other._rule \
                   and self._endOfMonth == other._endOfMonth \
                   and self._firstDate == other._firstDate \
                   and self._nextToLastDate == other._nextToLastDate



# -*- coding: utf-8 -*-


cdef class Deque:

    cdef size_t window
    cdef public bint is_full
    cdef double* con
    cdef public size_t start
    cdef public size_t count

    cdef double dump(self, double value, double default=*)
    cpdef void dumps(self, values)
    cdef inline size_t size(self)
    cdef inline bint isFull(self)
    cpdef size_t idx(self, double value)
    cpdef double sum(self)
    cdef void set_data(self, bytes data)


cpdef object rebuild(bytes data, size_t window, bint is_full, size_t start, size_t count)


# -*- coding: utf-8 -*-

from typing import Tuple
from typing import Union

import numpy as np

from .optimizers import LPOptimizer
from .optimizers import L1LPOptimizer



def linear_builder(er: np.ndarray,
                   lbound: Union[np.ndarray, float] = None,
                   ubound: Union[np.ndarray, float] = None,
                   risk_constraints: np.ndarray = None,
                   risk_target: Tuple[np.ndarray, np.ndarray] = None,
                   turn_over_target: float = None,
                   current_position: np.ndarray = None,
                   method: str = "deprecated") -> Tuple[str, np.ndarray, np.ndarray]:
    er = er.flatten()

    if risk_constraints is not None:
        risk_lbound = risk_target[0].reshape((-1, 1))
        risk_ubound = risk_target[1].reshape((-1, 1))
        cons_matrix = np.concatenate((risk_constraints.T, risk_lbound, risk_ubound), axis=1)
    else:
        cons_matrix = None

    if not turn_over_target or current_position is None:
        prob = LPOptimizer(-er, cons_matrix, lbound, ubound)

        if prob.status() == "optimal" or prob.status() == 'optimal_inaccurate':
            return prob.status(), prob.feval(), prob.x_value()
        else:
            raise (prob.status())
    elif turn_over_target:
        prob = L1LPOptimizer(objective=-er,
                             cons_matrix=cons_matrix,
                             current_pos=current_position,
                             target_turn_over=turn_over_target,
                             lbound=lbound,
                             ubound=ubound)

        if prob.status() == 'optimal' or prob.status() == 'optimal_inaccurate':
            return prob.status(), prob.feval(), prob.x_value()
        else:
            raise (prob.status())



# -*- coding: utf-8 -*-
import pdb
import datetime
import time
import json
from ultron.cluster.central.extern_modules.base_module import BaseModule
from ultron.cluster.central.extern_modules.middleware.module import ModelsSingleton

class Module(BaseModule):
    def __init__(self, name, redis_client):
        super(Module, self).__init__(name, redis_client)
        self._func = {'upload_packet':self.upload_packet}
     
    def process_respone(self, result):
        name = result['name']
        opcode = result['opcode']
        self._func[opcode](result)
    
    def upload_packet(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 + datetime.datetime.now().microsecond)
        queue_list = ModelsSingleton().get_all_queue()
        for queue in queue_list:
            self._redis_client.hset(queue,str(task_id),json.dumps(result))


# -*- coding: utf-8 -*-

import copy
import math
import numpy as np
cimport numpy as np
cimport cython
import six
from libc.math cimport isnan
from libc.math cimport log
from libc.math cimport fmax
from libc.math cimport fmin
from libc.math cimport sqrt
from ultron.sentry.Math.Accumulators.IAccumulators import build_holder
from ultron.sentry.Math.Accumulators.IAccumulators cimport Accumulator
from ultron.sentry.Math.Accumulators.IAccumulators cimport Pow
from ultron.sentry.Math.Accumulators.IAccumulators cimport Latest
from ultron.sentry.Math.MathConstants cimport NAN
import bisect


cdef class Diff(Accumulator):

    def __init__(self, x):
        super(Diff, self).__init__()
        self._curr = NAN
        self._previous = NAN
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        cdef double value
        self._inner.push(data)
        value = self._inner.result()
        if isnan(value):
            return NAN
        self._isFull = self._isFull or self._inner.isFull()
        self._previous = self._curr
        self._curr = value

    cpdef double result(self):
        return self._curr - self._previous

    def __str__(self):
        return "\\mathrm{{Diff}}({0})".format(str(self._inner))


cdef class SimpleReturn(Accumulator):

    def __init__(self, x):
        super(SimpleReturn, self).__init__()
        self._diff = NAN
        self._curr = NAN
        self._previous = NAN
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        cdef double value
        self._inner.push(data)
        value = self._inner.result()
        if isnan(value):
            return NAN
        self._isFull = self._isFull or self._inner.isFull()
        self._previous = self._curr
        self._curr = value

    @cython.cdivision(True)
    cpdef double result(self):
        cdef double denorm = self._previous
        if denorm:
            return self._curr / denorm - 1.
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{SimpleReturn}}({0})".format(str(self._inner))


cdef class LogReturn(Accumulator):

    def __init__(self, x):
        super(LogReturn, self).__init__()
        self._diff = NAN
        self._curr = NAN
        self._previous = NAN
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        cdef double value
        self._inner.push(data)
        value = self._inner.result()
        if isnan(value):
            return NAN
        self._isFull = self._isFull or self._inner.isFull()
        self._previous = self._curr
        self._curr = value

    @cython.cdivision(True)
    cpdef double result(self):
        cdef double denorm = self._previous
        if denorm:
            return log(self._curr / denorm)
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{LogReturn}}({0})".format(str(self._inner))


cdef class PositivePart(Accumulator):

    def __init__(self, x):
        super(PositivePart, self).__init__()
        self._pos = NAN
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        cdef double value
        self._inner.push(data)
        value = self._inner.result()
        if isnan(value):
            self._pos = NAN
        else:
            self._pos = fmax(value, 0.)
        self._isFull = self._isFull or self._inner.isFull()

    cpdef double result(self):
        return self._pos

    def __str__(self):
        return "\\mathrm{{PositivePart}}({0})".format(str(self._inner))


cdef class NegativePart(Accumulator):

    def __init__(self, x):
        super(NegativePart, self).__init__()
        self._neg = NAN
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        cdef double value
        self._inner.push(data)
        value = self._inner.result()
        if isnan(value):
            self._neg = NAN
        else:
            self._neg = fmin(value, 0.)
        self._isFull = self._isFull or self._inner.isFull()

    cpdef double result(self):
        return self._neg

    def __str__(self):
        return "\\mathrm{{NegativePart}}({0})".format(str(self._inner))


cdef class Max(Accumulator):

    def __init__(self, x):
        super(Max, self).__init__()
        self._currentMax = -np.inf
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        cdef double value
        self._inner.push(data)
        value = self._inner.result()
        if isnan(value):
            return NAN

        self._currentMax = fmax(value, self._currentMax)
        self._isFull = self._isFull or self._inner.isFull()

    cpdef double result(self):
        return self._currentMax

    def __str__(self):
        return "\\mathrm{{Max}}({0})".format(str(self._inner))

cdef class Maximum(Accumulator):

    def __init__(self, x, y):
        super(Maximum, self).__init__()
        self._currentMax = NAN
        self._isFull = False
        self._x = build_holder(x)
        self._y = build_holder(y)
        self._dependency = list(set(self._x.dependency + self._y.dependency))

    cpdef push(self, dict data):
        self._x.push(data)
        cdef double left = self._x.result()
        self._y.push(data)
        cdef double right = self._y.result()
        self._currentMax = fmax(left, right)
        self._isFull = self._isFull or (self._y.isFull() and self._y.isFull())

    cpdef double result(self):
        return self._currentMax

    def __str__(self):
        return "\\mathrm{{Maximum}}({0}, {1})".format(str(self._x), str(self._y))


cdef class Min(Accumulator):

    def __init__(self, x):
        super(Min, self).__init__()
        self._currentMin = np.inf
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        self._inner.push(data)
        cdef double value = self._inner.result()
        if isnan(value):
            return NAN

        self._currentMin = fmin(value, self._currentMin)
        self._isFull = self._isFull or self._inner.isFull()

    cpdef double result(self):
        return self._currentMin

    def __str__(self):
        return "\\mathrm{{Min}}({0})".format(str(self._inner))


cdef class Minimum(Accumulator):

    def __init__(self, x, y):
        super(Minimum, self).__init__()
        self._currentMin = NAN
        self._isFull = False
        self._x = build_holder(x)
        self._y = build_holder(y)
        self._dependency = list(set(self._x.dependency + self._y.dependency))

    cpdef push(self, dict data):
        self._x.push(data)
        cdef double left = self._x.result()
        self._y.push(data)
        cdef double right = self._y.result()
        self._currentMin = fmin(left, right)
        self._isFull = self._isFull or (self._y.isFull() and self._y.isFull())

    cpdef double result(self):
        return self._currentMin

    def __str__(self):
        return "\\mathrm{{Minimum}}({0}, {1})".format(str(self._x), str(self._y))


cdef class Sum(Accumulator):

    def __init__(self, x):
        super(Sum, self).__init__()
        self._currentSum = 0.
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        self._inner.push(data)
        cdef double value = self._inner.result()
        if isnan(value):
            return NAN

        self._currentSum += value
        self._isFull = self._isFull or self._inner.isFull()

    cpdef double result(self):
        return self._currentSum

    def __str__(self):
        return "\\mathrm{{Sum}}({0})".format(str(self._inner))


cdef class Average(Accumulator):

    def __init__(self, x):
        super(Average, self).__init__()
        self._currentSum = 0.
        self._currentCount = 0
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        self._inner.push(data)
        cdef double value = self._inner.result()
        if isnan(value):
            return NAN

        self._currentCount += 1
        self._currentSum += value
        self._isFull = self._isFull or self._inner.isFull()

    @cython.cdivision(True)
    cpdef double result(self):
        if self._currentCount:
            return self._currentSum / self._currentCount
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{Average}}({0})".format(str(self._inner))


cdef class XAverage(Accumulator):

    def __init__(self, window, x):
        super(XAverage, self).__init__()
        self._average = 0.0
        self._exp = 2.0 / (window + 1.)
        self._count = 0
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        self._inner.push(data)
        cdef double value = self._inner.result()
        if isnan(value):
            return NAN

        if self._count == 0:
            self._average = value
            self._count += 1
        else:
            self._average += self._exp * (value - self._average)
        self._isFull = self._isFull or self._inner.isFull()

    def __str__(self):
        return "\\mathrm{{XAverage}}({0}, {1})".format(2.0 / self._exp - 1., str(self._inner))

    cpdef double result(self):
        return self._average


cdef class Variance(Accumulator):

    def __init__(self, x, bint isPopulation=0):
        super(Variance, self).__init__()
        self._currentSum = 0.0
        self._currentSumSquare = 0.0
        self._currentCount = 0
        self._isPop = isPopulation
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        self._inner.push(data)
        cdef double value = self._inner.result()
        if isnan(value):
            return NAN

        self._isFull = True

        self._currentSum += value
        self._currentSumSquare += value * value
        self._currentCount += 1
        self._isFull = self._isFull or self._inner.isFull()

    @cython.cdivision(True)
    cpdef double result(self):

        cdef double tmp = self._currentSumSquare - self._currentSum * self._currentSum / self._currentCount
        cdef double pop_num = self._currentCount if self._isPop else self._currentCount - 1

        if pop_num:
            return tmp / pop_num
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{Variance}}({0})".format(str(self._inner))


cdef class Product(Accumulator):

    def __init__(self, x):
        super(Product, self).__init__()
        self._product = 1.0
        self._isFull = False
        self._inner = build_holder(x)
        self._window = self._inner.window
        self._dependency = self._inner.dependency

    cpdef push(self, dict data):
        self._inner.push(data)
        cdef double value = self._inner.result()
        if isnan(value):
            return NAN

        self._product *= value
        self._isFull = self._isFull or self._inner.isFull()

    cpdef double result(self):
        return self._product

    def __str__(self):
        return "\\mathrm{{Product}}({0})".format(str(self._inner))



# -*- coding: utf-8 -*-
import abc
import warnings
from distutils.version import LooseVersion
import arrow
import numpy as np
import pandas as pd
try:
    from sklearn import __version__ as sklearn_version
except ImportError:
    warnings.warn("not installed sklean,please pip install --upgrade sklearn")
try:
    from xgboost import __version__ as xgbboot_version
except ImportError:
    warnings.warn("not installed xgboost,please pip install --upgrade xgboost")
from ultron.factor.utilities import list_eq
from ultron.factor.utilities import encode
from ultron.factor.utilities import decode
from ultron.factor.data.transformer import Transformer

class ModelBase(metaclass=abc.ABCMeta):

    def __init__(self, features=None, fit_target=None):
        if features is not None:
            self.formulas = Transformer(features)
            self.features = self.formulas.names
        else:
            self.features = None

        if fit_target is not None:
            self.fit_target = Transformer(fit_target)
        else:
            self.fit_target = None
        self.impl = None
        self.trained_time = None

    def model_encode(self):
        return encode(self.impl)

    @classmethod
    def model_decode(cls, model_desc):
        return decode(model_desc)

    def __eq__(self, rhs):
        return self.model_encode() == rhs.model_encode() \
               and self.trained_time == rhs.trained_time \
               and list_eq(self.features, rhs.features) \
               and encode(self.formulas) == encode(rhs.formulas) \
               and encode(self.fit_target) == encode(rhs.fit_target)

    def fit(self, x: pd.DataFrame, y: np.ndarray):
        self.impl.fit(x[self.features].values, y.flatten())
        self.trained_time = arrow.now().format("YYYY-MM-DD HH:mm:ss")

    def predict(self, x: pd.DataFrame) -> np.ndarray:
        return self.impl.predict(x[self.features].values)

    def score(self, x: pd.DataFrame, y: np.ndarray) -> float:
        return self.impl.score(x[self.features].values, y)

    def ic(self, x: pd.DataFrame, y: np.ndarray) -> float:
        predict_y = self.impl.predict(x[self.features].values)
        return np.corrcoef(predict_y, y)[0, 1]

    @abc.abstractmethod
    def save(self) -> dict:

        if self.__class__.__module__ == '__main__':
            warnings.warn("model is defined in a main module. The model_name may not be correct.")

        model_desc = dict(model_name=self.__class__.__module__ + "." + self.__class__.__name__,
                          language='python',
                          saved_time=arrow.now().format("YYYY-MM-DD HH:mm:ss"),
                          features=list(self.features),
                          trained_time=self.trained_time,
                          desc=self.model_encode(),
                          formulas=encode(self.formulas),
                          fit_target=encode(self.fit_target),
                          internal_model=self.impl.__class__.__module__ + "." + self.impl.__class__.__name__)
        return model_desc

    @classmethod
    @abc.abstractmethod
    def load(cls, model_desc: dict):
        obj_layout = cls()
        obj_layout.features = model_desc['features']
        obj_layout.formulas = decode(model_desc['formulas'])
        obj_layout.trained_time = model_desc['trained_time']
        obj_layout.impl = cls.model_decode(model_desc['desc'])
        if 'fit_target' in model_desc:
            obj_layout.fit_target = decode(model_desc['fit_target'])
        else:
            obj_layout.fit_target = None
        return obj_layout


def create_model_base(party_name=None):
    if not party_name:
        return ModelBase
    else:
        class ExternalLibBase(ModelBase):
            _lib_name = party_name

            def save(self) -> dict:
                model_desc = super().save()
                if self._lib_name == 'sklearn':
                    model_desc[self._lib_name + "_version"] = sklearn_version
                elif self._lib_name == 'xgboost':
                    model_desc[self._lib_name + "_version"] = xgbboot_version
                else:
                    raise ValueError("3rd party lib name ({0}) is not recognized".format(self._lib_name))
                return model_desc

            @classmethod
            def load(cls, model_desc: dict):
                obj_layout = super().load(model_desc)

                if cls._lib_name == 'sklearn':
                    current_version = sklearn_version
                elif cls._lib_name == 'xgboost':
                    current_version = xgbboot_version
                else:
                    raise ValueError("3rd party lib name ({0}) is not recognized".format(cls._lib_name))

                if LooseVersion(current_version) < LooseVersion(model_desc[cls._lib_name + "_version"]):
                    warnings.warn('Current {2} version {0} is lower than the model version {1}. '
                                         'Loaded model may work incorrectly.'.format(sklearn_version,
                                                                                     model_desc[cls._lib_name],
                                                                                     cls._lib_name))
                return obj_layout
        return ExternalLibBase


# -*- coding: utf-8 -*-
import pdb
import datetime
import time
import json
from ultron.cluster.central.extern_modules.base_module import BaseModule
from ultron.cluster.central.extern_modules.middleware.module import ModelsSingleton

class Module(BaseModule):
    def __init__(self, name, redis_client):
        super(Module, self).__init__(name, redis_client)
        self._func = {'startup_task':self.startup_task,
                     'shutoff_task':self.shutoff_task,
                     'restart_task':self.restart_task,
                     'update_task':self.update_task}
        
    
    def process_respone(self, result):
        name = result['name']
        opcode = result['opcode']
        self._func[opcode](result)
    
    def update_task(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 + datetime.datetime.now().microsecond)
        wid = result['wid']
        #获取消息通道
        queue = ModelsSingleton().get_queue(wid)
        self._redis_client.hset(queue, str(task_id),json.dumps(result))
                      
    def startup_task(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 + datetime.datetime.now().microsecond)
        work_name = result['work_name']
        dir_name = result['dir_name']
        task_info = {'name':'tasks','opcode':'startup_task','work_name':work_name,
                    'dir_name':dir_name}
        queue_list = ModelsSingleton().get_all_queue()
        for queue in queue_list:
            self._redis_client.hset(queue,str(task_id),json.dumps(task_info))
    
    def shutoff_task(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 + datetime.datetime.now().microsecond)
        work_name = result['work_name']
        dir_name = result['dir_name']
        task_info = {'name':'tasks','opcode':'shutoff_task','work_name':work_name,
                    'dir_name':dir_name}
        queue_list = ModelsSingleton().get_all_queue()
        for queue in queue_list:
            self._redis_client.hset(queue,str(task_id),json.dumps(task_info))
        
    def restart_task(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 + datetime.datetime.now().microsecond)
        work_name = result['work_name']
        dir_name = result['dir_name']
        task_info = {'name':'tasks','opcode':'restart_task','work_name':work_name,
                    'dir_name':dir_name}
        queue_list = ModelsSingleton().get_all_queue()
        for queue in queue_list:
            self._redis_client.hset(queue,str(task_id),json.dumps(task_info))


# -*- coding: utf-8 -*-

from ultron.tradingday.Enums._Weekdays cimport Weekdays
from ultron.tradingday.Enums._TimeUnits cimport TimeUnits
from ultron.tradingday.Enums._Months cimport Months
from ultron.tradingday.Enums._BizDayConventions cimport BizDayConventions
from ultron.tradingday.Date cimport Date
from ultron.tradingday.Period cimport Period

cdef class Calendar(object):
    def __init__(self, str holCenter):
        holCenter = holCenter.lower()
        try:
            self._impl = _holDict[holCenter]()
        except KeyError:
            raise ValueError("{0} is not a valid description of a holiday center".format(holCenter))
        self.name = holCenter

    cpdef isBizDay(self, Date d):
        return self._impl.isBizDay(d)

    cpdef isHoliday(self, Date d):
        return not self._impl.isBizDay(d)

    cpdef isWeekEnd(self, int weekday):
        return self._impl.isWeekEnd(weekday)

    cpdef isEndOfMonth(self, Date d):
        return d.month() != self.adjustDate(d + 1).month()

    cpdef endOfMonth(self, Date d):
        return self.adjustDate(Date.endOfMonth(d), BizDayConventions.Preceding)

    cpdef bizDaysBetween(self, Date fromDate, Date toDate, bint includeFirst=True, bint includeLast=False):
        cdef int wd = 0
        cdef Date d

        if fromDate != toDate:
            if fromDate < toDate:
                d = fromDate
                while d < toDate:
                    if self.isBizDay(d):
                        wd += 1
                    d += 1
                if self.isBizDay(toDate):
                    wd += 1
            elif fromDate > toDate:
                d = toDate
                while d < fromDate:
                    if self.isBizDay(d):
                        wd += 1
                    d += 1
                if self.isBizDay(fromDate):
                    wd += 1
            if self.isBizDay(fromDate) and not includeFirst:
                wd -= 1
            if self.isBizDay(toDate) and not includeLast:
                wd -= 1
        return wd

    cpdef adjustDate(self, Date d, int c=BizDayConventions.Following):

        cdef Date d1
        cdef Date d2

        if c == BizDayConventions.Unadjusted:
            return d
        d1 = d

        if c == BizDayConventions.Following or c == BizDayConventions.ModifiedFollowing or \
                c == BizDayConventions.HalfMonthModifiedFollowing:
            while self.isHoliday(d1):
                d1 += 1
            if c == BizDayConventions.ModifiedFollowing or c == BizDayConventions.HalfMonthModifiedFollowing:
                if d1.month() != d.month():
                    return self.adjustDate(d, BizDayConventions.Preceding)
                if c == BizDayConventions.HalfMonthModifiedFollowing:
                    if d.dayOfMonth() <= 15 < d1.dayOfMonth():
                        return self.adjustDate(d, BizDayConventions.Preceding)
        elif c == BizDayConventions.Preceding or c == BizDayConventions.ModifiedPreceding:
            while self.isHoliday(d1):
                d1 -= 1
            if c == BizDayConventions.ModifiedPreceding and d1.month() != d.month():
                return self.adjustDate(d, BizDayConventions.Following)
        elif c == BizDayConventions.Nearest:
            d2 = d
            while self.isHoliday(d1) and self.isHoliday(d2):
                d1 += 1
                d2 -= 1

            if self.isHoliday(d1):
                return d2
            else:
                return d1
        else:
            raise ValueError("unknown business-day convention")
        return d1

    cpdef advanceDate(self, Date d, Period period, int c=BizDayConventions.Following, bint endOfMonth=False):

        cdef int n
        cdef int units
        cdef Date d1

        n = period.length()
        units = period.units()

        if n == 0:
            return self.adjustDate(d, c)
        elif units == TimeUnits.BDays:
            d1 = d
            if n > 0:
                while n > 0:
                    d1 += 1
                    while self.isHoliday(d1):
                        d1 += 1
                    n -= 1
            else:
                while n < 0:
                    d1 -= 1
                    while self.isHoliday(d1):
                        d1 -= 1
                    n += 1
            return d1
        elif units == TimeUnits.Days or units == TimeUnits.Weeks:
            d1 = d + period
            return self.adjustDate(d1, c)
        else:
            d1 = d + period
            if endOfMonth and self.isEndOfMonth(d):
                return self.endOfMonth(d1)
            return self.adjustDate(d1, c)

    cpdef holDatesList(self, Date fromDate, Date toDate, bint includeWeekEnds=True):
        cdef list result = []
        cdef Date d = fromDate

        while d <= toDate:
            if self.isHoliday(d) and (includeWeekEnds or not self.isWeekEnd(d.weekday())):
                result.append(d)
            d += 1
        return result

    cpdef bizDatesList(self, Date fromDate, Date toDate):
        cdef list result = []
        cdef Date d = fromDate

        while d <= toDate:
            if self.isBizDay(d):
                result.append(d)
            d += 1
        return result

    def __richcmp__(self, right, int op):
        if op == 2:
            return self._impl == right._impl

    def __deepcopy__(self, memo):
        return Calendar(self.name)

    def __reduce__(self):
        d = {}

        return Calendar, (self.name,), d

    def __setstate__(self, state):
        pass

cdef class CalendarImpl(object):
    cdef bint isBizDay(self, Date date):
        pass

    cdef bint isWeekEnd(self, int weekDay):
        pass

cdef set sse_holDays = {
    Date(1991, 1, 1),
    Date(1991, 2, 15),
    Date(1991, 2, 18),
    Date(1991, 5, 1),
    Date(1991, 10, 1),
    Date(1991, 10, 2),
    Date(1992, 1, 1),
    Date(1992, 2, 4),
    Date(1992, 2, 5),
    Date(1992, 2, 6),
    Date(1992, 5, 1),
    Date(1992, 10, 1),
    Date(1992, 10, 2),
    Date(1993, 1, 1),
    Date(1993, 1, 25),
    Date(1993, 1, 26),
    Date(1993, 10, 1),
    Date(1994, 2, 7),
    Date(1994, 2, 8),
    Date(1994, 2, 9),
    Date(1994, 2, 10),
    Date(1994, 2, 11),
    Date(1994, 5, 2),
    Date(1994, 10, 3),
    Date(1994, 10, 4),
    Date(1995, 1, 2),
    Date(1995, 1, 30),
    Date(1995, 1, 31),
    Date(1995, 2, 1),
    Date(1995, 2, 2),
    Date(1995, 2, 3),
    Date(1995, 5, 1),
    Date(1995, 10, 2),
    Date(1995, 10, 3),
    Date(1996, 1, 1),
    Date(1996, 2, 19),
    Date(1996, 2, 20),
    Date(1996, 2, 21),
    Date(1996, 2, 22),
    Date(1996, 2, 23),
    Date(1996, 2, 26),
    Date(1996, 2, 27),
    Date(1996, 2, 28),
    Date(1996, 2, 29),
    Date(1996, 3, 1),
    Date(1996, 5, 1),
    Date(1996, 9, 30),
    Date(1996, 10, 1),
    Date(1996, 10, 2),
    Date(1997, 1, 1),
    Date(1997, 2, 3),
    Date(1997, 2, 4),
    Date(1997, 2, 5),
    Date(1997, 2, 6),
    Date(1997, 2, 7),
    Date(1997, 2, 10),
    Date(1997, 2, 11),
    Date(1997, 2, 12),
    Date(1997, 2, 13),
    Date(1997, 2, 14),
    Date(1997, 5, 1),
    Date(1997, 5, 2),
    Date(1997, 6, 30),
    Date(1997, 7, 1),
    Date(1997, 10, 1),
    Date(1997, 10, 2),
    Date(1997, 10, 3),
    Date(1998, 1, 1),
    Date(1998, 1, 2),
    Date(1998, 1, 26),
    Date(1998, 1, 27),
    Date(1998, 1, 28),
    Date(1998, 1, 29),
    Date(1998, 1, 30),
    Date(1998, 2, 2),
    Date(1998, 2, 3),
    Date(1998, 2, 4),
    Date(1998, 2, 5),
    Date(1998, 2, 6),
    Date(1998, 5, 1),
    Date(1998, 10, 1),
    Date(1998, 10, 2),
    Date(1999, 1, 1),
    Date(1999, 2, 10),
    Date(1999, 2, 11),
    Date(1999, 2, 12),
    Date(1999, 2, 15),
    Date(1999, 2, 16),
    Date(1999, 2, 17),
    Date(1999, 2, 18),
    Date(1999, 2, 19),
    Date(1999, 2, 22),
    Date(1999, 2, 23),
    Date(1999, 2, 24),
    Date(1999, 2, 25),
    Date(1999, 2, 26),
    Date(1999, 5, 3),
    Date(1999, 10, 1),
    Date(1999, 10, 4),
    Date(1999, 10, 5),
    Date(1999, 10, 6),
    Date(1999, 10, 7),
    Date(1999, 12, 20),
    Date(1999, 12, 31),
    Date(2000, 1, 3),
    Date(2000, 1, 31),
    Date(2000, 2, 1),
    Date(2000, 2, 2),
    Date(2000, 2, 3),
    Date(2000, 2, 4),
    Date(2000, 2, 7),
    Date(2000, 2, 8),
    Date(2000, 2, 9),
    Date(2000, 2, 10),
    Date(2000, 2, 11),
    Date(2000, 5, 1),
    Date(2000, 5, 2),
    Date(2000, 5, 3),
    Date(2000, 5, 4),
    Date(2000, 5, 5),
    Date(2000, 10, 2),
    Date(2000, 10, 3),
    Date(2000, 10, 4),
    Date(2000, 10, 5),
    Date(2000, 10, 6),
    Date(2001, 1, 1),
    Date(2001, 1, 22),
    Date(2001, 1, 23),
    Date(2001, 1, 24),
    Date(2001, 1, 25),
    Date(2001, 1, 26),
    Date(2001, 1, 29),
    Date(2001, 1, 30),
    Date(2001, 1, 31),
    Date(2001, 2, 1),
    Date(2001, 2, 2),
    Date(2001, 5, 1),
    Date(2001, 5, 2),
    Date(2001, 5, 3),
    Date(2001, 5, 4),
    Date(2001, 5, 7),
    Date(2001, 10, 1),
    Date(2001, 10, 2),
    Date(2001, 10, 3),
    Date(2001, 10, 4),
    Date(2001, 10, 5),
    Date(2002, 1, 1),
    Date(2002, 1, 2),
    Date(2002, 1, 3),
    Date(2002, 2, 11),
    Date(2002, 2, 12),
    Date(2002, 2, 13),
    Date(2002, 2, 14),
    Date(2002, 2, 15),
    Date(2002, 2, 18),
    Date(2002, 2, 19),
    Date(2002, 2, 20),
    Date(2002, 2, 21),
    Date(2002, 2, 22),
    Date(2002, 5, 1),
    Date(2002, 5, 2),
    Date(2002, 5, 3),
    Date(2002, 5, 6),
    Date(2002, 5, 7),
    Date(2002, 9, 30),
    Date(2002, 10, 1),
    Date(2002, 10, 2),
    Date(2002, 10, 3),
    Date(2002, 10, 4),
    Date(2002, 10, 7),
    Date(2003, 1, 1),
    Date(2003, 1, 30),
    Date(2003, 1, 31),
    Date(2003, 2, 3),
    Date(2003, 2, 4),
    Date(2003, 2, 5),
    Date(2003, 2, 6),
    Date(2003, 2, 7),
    Date(2003, 5, 1),
    Date(2003, 5, 2),
    Date(2003, 5, 5),
    Date(2003, 5, 6),
    Date(2003, 5, 7),
    Date(2003, 5, 8),
    Date(2003, 5, 9),
    Date(2003, 10, 1),
    Date(2003, 10, 2),
    Date(2003, 10, 3),
    Date(2003, 10, 6),
    Date(2003, 10, 7),
    Date(2004, 1, 1),
    Date(2004, 1, 19),
    Date(2004, 1, 20),
    Date(2004, 1, 21),
    Date(2004, 1, 22),
    Date(2004, 1, 23),
    Date(2004, 1, 26),
    Date(2004, 1, 27),
    Date(2004, 1, 28),
    Date(2004, 5, 3),
    Date(2004, 5, 4),
    Date(2004, 5, 5),
    Date(2004, 5, 6),
    Date(2004, 5, 7),
    Date(2004, 10, 1),
    Date(2004, 10, 4),
    Date(2004, 10, 5),
    Date(2004, 10, 6),
    Date(2004, 10, 7),
    Date(2005, 1, 3),
    Date(2005, 2, 7),
    Date(2005, 2, 8),
    Date(2005, 2, 9),
    Date(2005, 2, 10),
    Date(2005, 2, 11),
    Date(2005, 2, 14),
    Date(2005, 2, 15),
    Date(2005, 5, 2),
    Date(2005, 5, 3),
    Date(2005, 5, 4),
    Date(2005, 5, 5),
    Date(2005, 5, 6),
    Date(2005, 10, 3),
    Date(2005, 10, 4),
    Date(2005, 10, 5),
    Date(2005, 10, 6),
    Date(2005, 10, 7),
    Date(2006, 1, 2),
    Date(2006, 1, 3),
    Date(2006, 1, 26),
    Date(2006, 1, 27),
    Date(2006, 1, 30),
    Date(2006, 1, 31),
    Date(2006, 2, 1),
    Date(2006, 2, 2),
    Date(2006, 2, 3),
    Date(2006, 5, 1),
    Date(2006, 5, 2),
    Date(2006, 5, 3),
    Date(2006, 5, 4),
    Date(2006, 5, 5),
    Date(2006, 10, 2),
    Date(2006, 10, 3),
    Date(2006, 10, 4),
    Date(2006, 10, 5),
    Date(2006, 10, 6),
    Date(2007, 1, 1),
    Date(2007, 1, 2),
    Date(2007, 1, 3),
    Date(2007, 2, 19),
    Date(2007, 2, 20),
    Date(2007, 2, 21),
    Date(2007, 2, 22),
    Date(2007, 2, 23),
    Date(2007, 5, 1),
    Date(2007, 5, 2),
    Date(2007, 5, 3),
    Date(2007, 5, 4),
    Date(2007, 5, 7),
    Date(2007, 10, 1),
    Date(2007, 10, 2),
    Date(2007, 10, 3),
    Date(2007, 10, 4),
    Date(2007, 10, 5),
    Date(2007, 12, 31),
    Date(2008, 1, 1),
    Date(2008, 2, 6),
    Date(2008, 2, 7),
    Date(2008, 2, 8),
    Date(2008, 2, 11),
    Date(2008, 2, 12),
    Date(2008, 4, 4),
    Date(2008, 5, 1),
    Date(2008, 5, 2),
    Date(2008, 6, 9),
    Date(2008, 9, 15),
    Date(2008, 9, 29),
    Date(2008, 9, 30),
    Date(2008, 10, 1),
    Date(2008, 10, 2),
    Date(2008, 10, 3),
    Date(2009, 1, 1),
    Date(2009, 1, 2),
    Date(2009, 1, 26),
    Date(2009, 1, 27),
    Date(2009, 1, 28),
    Date(2009, 1, 29),
    Date(2009, 1, 30),
    Date(2009, 4, 6),
    Date(2009, 5, 1),
    Date(2009, 5, 28),
    Date(2009, 5, 29),
    Date(2009, 10, 1),
    Date(2009, 10, 2),
    Date(2009, 10, 5),
    Date(2009, 10, 6),
    Date(2009, 10, 7),
    Date(2009, 10, 8),
    Date(2010, 1, 1),
    Date(2010, 2, 15),
    Date(2010, 2, 16),
    Date(2010, 2, 17),
    Date(2010, 2, 18),
    Date(2010, 2, 19),
    Date(2010, 4, 5),
    Date(2010, 5, 3),
    Date(2010, 6, 14),
    Date(2010, 6, 15),
    Date(2010, 6, 16),
    Date(2010, 9, 22),
    Date(2010, 9, 23),
    Date(2010, 9, 24),
    Date(2010, 10, 1),
    Date(2010, 10, 4),
    Date(2010, 10, 5),
    Date(2010, 10, 6),
    Date(2010, 10, 7),
    Date(2011, 1, 3),
    Date(2011, 2, 2),
    Date(2011, 2, 3),
    Date(2011, 2, 4),
    Date(2011, 2, 7),
    Date(2011, 2, 8),
    Date(2011, 4, 4),
    Date(2011, 4, 5),
    Date(2011, 5, 2),
    Date(2011, 6, 6),
    Date(2011, 9, 12),
    Date(2011, 10, 3),
    Date(2011, 10, 4),
    Date(2011, 10, 5),
    Date(2011, 10, 6),
    Date(2011, 10, 7),
    Date(2012, 1, 2),
    Date(2012, 1, 3),
    Date(2012, 1, 23),
    Date(2012, 1, 24),
    Date(2012, 1, 25),
    Date(2012, 1, 26),
    Date(2012, 1, 27),
    Date(2012, 4, 2),
    Date(2012, 4, 3),
    Date(2012, 4, 4),
    Date(2012, 4, 30),
    Date(2012, 5, 1),
    Date(2012, 6, 22),
    Date(2012, 10, 1),
    Date(2012, 10, 2),
    Date(2012, 10, 3),
    Date(2012, 10, 4),
    Date(2012, 10, 5),
    Date(2013, 1, 1),
    Date(2013, 1, 2),
    Date(2013, 1, 3),
    Date(2013, 2, 11),
    Date(2013, 2, 12),
    Date(2013, 2, 13),
    Date(2013, 2, 14),
    Date(2013, 2, 15),
    Date(2013, 4, 4),
    Date(2013, 4, 5),
    Date(2013, 4, 29),
    Date(2013, 4, 30),
    Date(2013, 5, 1),
    Date(2013, 6, 10),
    Date(2013, 6, 11),
    Date(2013, 6, 12),
    Date(2013, 9, 19),
    Date(2013, 9, 20),
    Date(2013, 10, 1),
    Date(2013, 10, 2),
    Date(2013, 10, 3),
    Date(2013, 10, 4),
    Date(2013, 10, 7),
    Date(2014, 1, 1),
    Date(2014, 1, 31),
    Date(2014, 2, 3),
    Date(2014, 2, 4),
    Date(2014, 2, 5),
    Date(2014, 2, 6),
    Date(2014, 4, 7),
    Date(2014, 5, 1),
    Date(2014, 5, 2),
    Date(2014, 6, 2),
    Date(2014, 9, 8),
    Date(2014, 10, 1),
    Date(2014, 10, 2),
    Date(2014, 10, 3),
    Date(2014, 10, 6),
    Date(2014, 10, 7),
    Date(2015, 1, 1),
    Date(2015, 1, 2),
    Date(2015, 2, 18),
    Date(2015, 2, 19),
    Date(2015, 2, 20),
    Date(2015, 2, 23),
    Date(2015, 2, 24),
    Date(2015, 4, 6),
    Date(2015, 5, 1),
    Date(2015, 6, 22),
    Date(2015, 9, 3),
    Date(2015, 9, 4),
    Date(2015, 10, 1),
    Date(2015, 10, 2),
    Date(2015, 10, 5),
    Date(2015, 10, 6),
    Date(2015, 10, 7),
    Date(2016, 1, 1),
    Date(2016, 2, 8),
    Date(2016, 2, 9),
    Date(2016, 2, 10),
    Date(2016, 2, 11),
    Date(2016, 2, 12),
    Date(2016, 4, 4),
    Date(2016, 5, 1),
    Date(2016, 5, 2),
    Date(2016, 6, 9),
    Date(2016, 6, 10),
    Date(2016, 9, 15),
    Date(2016, 9, 16),
    Date(2016, 10, 3),
    Date(2016, 10, 4),
    Date(2016, 10, 5),
    Date(2016, 10, 6),
    Date(2016, 10, 7),
    Date(2017, 1, 2),
    Date(2017, 1, 27),
    Date(2017, 1, 30),
    Date(2017, 1, 31),
    Date(2017, 2, 1),
    Date(2017, 2, 2),
    Date(2017, 4, 3),
    Date(2017, 4, 4),
    Date(2017, 5, 1),
    Date(2017, 5, 29),
    Date(2017, 5, 30),
    Date(2017, 10, 2),
    Date(2017, 10, 3),
    Date(2017, 10, 4),
    Date(2017, 10, 5),
    Date(2017, 10, 6),
    Date(2018, 1, 1),
    Date(2018, 2, 15),
    Date(2018, 2, 16),
    Date(2018, 2, 19),
    Date(2018, 2, 20),
    Date(2018, 2, 21),
    Date(2018, 4, 5),
    Date(2018, 4, 6),
    Date(2018, 4, 30),
    Date(2018, 5, 1),
    Date(2018, 6, 18),
    Date(2018, 9, 24),
    Date(2018, 10, 1),
    Date(2018, 10, 2),
    Date(2018, 10, 3),
    Date(2018, 10, 4),
    Date(2018, 10, 5),
    Date(2018, 12, 31),
    Date(2019, 1, 1),
    Date(2019, 2, 4),
    Date(2019, 2, 5),
    Date(2019, 2, 6),
    Date(2019, 2, 7),
    Date(2019, 2, 8),
    Date(2019, 4, 5),
    Date(2019, 5, 1),
    Date(2019, 5, 2),
    Date(2019, 5, 3),
    Date(2019, 6, 7),
    Date(2019, 9, 13),
    Date(2019, 10, 1),
    Date(2019, 10, 2),
    Date(2019, 10, 3),
    Date(2019, 10, 4),
    Date(2019, 10, 7),
    Date(2020, 1, 1),
    Date(2020, 1, 24),
    Date(2020, 1, 27),
    Date(2020, 1, 28),
    Date(2020, 1, 29),
    Date(2020, 1, 30),
    Date(2020, 1, 31),
    Date(2020, 4, 6),
    Date(2020, 5, 1),
    Date(2020, 5, 4),
    Date(2020, 5, 5),
    Date(2020, 6, 25),
    Date(2020, 6, 26),
    Date(2020, 10, 1),
    Date(2020, 10, 2),
    Date(2020, 10, 5),
    Date(2020, 10, 6),
    Date(2020, 10, 7),
    Date(2020, 10, 8),
    Date(2021, 1, 1),
    Date(2021, 2, 11),
    Date(2021, 2, 12),
    Date(2021, 2, 15),
    Date(2021, 2, 16),
    Date(2021, 2, 17),
    Date(2021, 4, 5),
    Date(2021, 5, 3),
    Date(2021, 5, 4),
    Date(2021, 5, 5),
    Date(2021, 6, 14),
    Date(2021, 9, 20),
    Date(2021, 9, 21),
    Date(2021, 10, 1),
    Date(2021, 10, 4),
    Date(2021, 10, 5),
    Date(2021, 10, 6),
    Date(2021, 10, 7),
    Date(2022, 1, 3),
    Date(2022, 1, 31),
    Date(2022, 2, 1),
    Date(2022, 2, 2),
    Date(2022, 2, 3),
    Date(2022, 2, 4),
    Date(2022, 4, 4),
    Date(2022, 4, 5),
    Date(2022, 5, 2),
    Date(2022, 5, 3),
    Date(2022, 5, 4),
    Date(2022, 6, 3),
    Date(2022, 9, 12),
    Date(2022, 10, 3),
    Date(2022, 10, 4),
    Date(2022, 10, 5),
    Date(2022, 10, 6),
    Date(2022, 10, 7)
}


cdef set sse_working_weekends = {
    # 1992
    Date.westernStyle(4, Months.October, 1992)
}


cdef class ChinaSseImpl(CalendarImpl):
    def __init__(self):
        pass

    cdef bint isBizDay(self, Date date):
        cdef int w = date.weekday()
        if (self.isWeekEnd(w) or date in sse_holDays) and date not in sse_working_weekends:
            return False
        return True

    cdef bint isWeekEnd(self, int weekDay):
        return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday

    def __richcmp__(self, right, int op):
        if op == 2:
            return isinstance(right, ChinaSseImpl)

cdef set ib_working_weekends = {
    # 1992
    Date.westernStyle(4, Months.October, 1992),
    # 2005
    Date.westernStyle(5, Months.February, 2005),
    Date.westernStyle(6, Months.February, 2005),
    Date.westernStyle(30, Months.April, 2005),
    Date.westernStyle(8, Months.May, 2005),
    Date.westernStyle(8, Months.October, 2005),
    Date.westernStyle(9, Months.October, 2005),
    Date.westernStyle(31, Months.December, 2005),
    # 2006
    Date.westernStyle(28, Months.January, 2006),
    Date.westernStyle(29, Months.April, 2006),
    Date.westernStyle(30, Months.April, 2006),
    Date.westernStyle(30, Months.September, 2006),
    Date.westernStyle(30, Months.December, 2006),
    Date.westernStyle(31, Months.December, 2006),
    # 2007
    Date.westernStyle(17, Months.February, 2007),
    Date.westernStyle(25, Months.February, 2007),
    Date.westernStyle(28, Months.April, 2007),
    Date.westernStyle(29, Months.April, 2007),
    Date.westernStyle(29, Months.September, 2007),
    Date.westernStyle(30, Months.September, 2007),
    Date.westernStyle(29, Months.December, 2007),
    # 2008
    Date.westernStyle(2, Months.February, 2008),
    Date.westernStyle(3, Months.February, 2008),
    Date.westernStyle(4, Months.May, 2008),
    Date.westernStyle(27, Months.September, 2008),
    Date.westernStyle(28, Months.September, 2008),
    # 2009
    Date.westernStyle(4, Months.January, 2009),
    Date.westernStyle(24, Months.January, 2009),
    Date.westernStyle(1, Months.February, 2009),
    Date.westernStyle(31, Months.May, 2009),
    Date.westernStyle(27, Months.September, 2009),
    Date.westernStyle(10, Months.October, 2009),
    # 2010
    Date.westernStyle(20, Months.February, 2010),
    Date.westernStyle(21, Months.February, 2010),
    Date.westernStyle(12, Months.June, 2010),
    Date.westernStyle(13, Months.June, 2010),
    Date.westernStyle(19, Months.September, 2010),
    Date.westernStyle(25, Months.September, 2010),
    Date.westernStyle(26, Months.September, 2010),
    Date.westernStyle(9, Months.October, 2010),
    # 2011
    Date.westernStyle(30, Months.January, 2011),
    Date.westernStyle(12, Months.February, 2011),
    Date.westernStyle(2, Months.April, 2011),
    Date.westernStyle(8, Months.October, 2011),
    Date.westernStyle(9, Months.October, 2011),
    Date.westernStyle(31, Months.December, 2011),
    # 2012
    Date.westernStyle(21, Months.January, 2012),
    Date.westernStyle(29, Months.January, 2012),
    Date.westernStyle(31, Months.March, 2012),
    Date.westernStyle(1, Months.April, 2012),
    Date.westernStyle(28, Months.April, 2012),
    Date.westernStyle(29, Months.September, 2012),
    # 2013
    Date.westernStyle(5, Months.January, 2013),
    Date.westernStyle(6, Months.January, 2013),
    Date.westernStyle(16, Months.February, 2013),
    Date.westernStyle(17, Months.February, 2013),
    Date.westernStyle(7, Months.April, 2013),
    Date.westernStyle(27, Months.April, 2013),
    Date.westernStyle(28, Months.April, 2013),
    Date.westernStyle(8, Months.June, 2013),
    Date.westernStyle(9, Months.June, 2013),
    Date.westernStyle(22, Months.September, 2013),
    Date.westernStyle(29, Months.September, 2013),
    Date.westernStyle(12, Months.October, 2013),
    # 2014
    Date.westernStyle(26, Months.January, 2014),
    Date.westernStyle(8, Months.February, 2014),
    Date.westernStyle(4, Months.May, 2014),
    Date.westernStyle(28, Months.September, 2014),
    Date.westernStyle(11, Months.October, 2014),
    # 2015
    Date.westernStyle(4, Months.January, 2015),
    Date.westernStyle(15, Months.February, 2015),
    Date.westernStyle(28, Months.February, 2015),
    Date.westernStyle(6, Months.September, 2015),
    Date.westernStyle(10, Months.October, 2015),
    # 2016
    Date.westernStyle(6, Months.February, 2016),
    Date.westernStyle(14, Months.February, 2016),
    Date.westernStyle(12, Months.June, 2016),
    Date.westernStyle(18, Months.September, 2016),
    Date.westernStyle(8, Months.October, 2016),
    Date.westernStyle(9, Months.October, 2016),
    # 2017
    Date.westernStyle(22, Months.January, 2017),
    Date.westernStyle(4, Months.February, 2017),
    Date.westernStyle(1, Months.April, 2017),
    Date.westernStyle(27, Months.May, 2017),
    Date.westernStyle(30, Months.September, 2017),
    # 2018
    Date.westernStyle(11, Months.February, 2018),
    Date.westernStyle(24, Months.February, 2018),
    Date.westernStyle(8, Months.April, 2018),
    Date.westernStyle(28, Months.April, 2018),
    Date.westernStyle(29, Months.September, 2018),
    Date.westernStyle(30, Months.September, 2018),
    Date.westernStyle(29, Months.December, 2018),
    # 2019
    Date.westernStyle(2, Months.February, 2019),
    Date.westernStyle(3, Months.February, 2019),
    Date.westernStyle(28, Months.April, 2019),
    Date.westernStyle(5, Months.May, 2019),
    Date.westernStyle(29, Months.September, 2019),
    Date.westernStyle(12, Months.October, 2019),
    # 2020
    Date.westernStyle(19, Months.January, 2020),
    Date.westernStyle(26, Months.April, 2020),
    Date.westernStyle(9, Months.May, 2020),
    Date.westernStyle(28, Months.June, 2020),
    Date.westernStyle(27, Months.September, 2020),
    Date.westernStyle(10, Months.October, 2020),
    # 2021
    Date(2021, 2, 7),
    Date(2021, 2, 20),
    Date(2021, 4, 25),
    Date(2021, 5, 8),
    Date(2021, 9, 18),
    Date(2021, 9, 26),
    Date(2021, 10, 9),
    # 2022
    Date(2022, 1, 29),
    Date(2022, 1, 30),
    Date(2022, 4, 2),
    Date(2022, 4, 24),
    Date(2022, 5, 7),
    Date(2022, 10, 8),
    Date(2022, 10, 9)
}

cdef ChinaSseImpl _sseImpl = ChinaSseImpl()

cdef class ChinaIBImpl(CalendarImpl):
    def __init__(self):
        pass

    cpdef bint isBizDay(self, Date date):
        return _sseImpl.isBizDay(date) or date in ib_working_weekends

    cpdef bint isWeekEnd(self, int weekDay):
        return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday

    def __richcmp__(self, right, int op):
        if op == 2:
            return isinstance(right, ChinaIBImpl)

cdef class NullCalendar(CalendarImpl):
    def __init__(self):
        pass

    cdef bint isBizDay(self, Date date):
        return True

    cdef bint isWeekEnd(self, int weekDay):
        return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday

    def __richcmp__(self, right, int op):
        if op == 2:
            return isinstance(right, NullCalendar)

cdef class ChinaCFFEXImpl(CalendarImpl):
    def __init__(self):
        pass

    cdef bint isBizDay(self, Date date):
        return _sseImpl.isBizDay(date)

    cdef bint isWeekEnd(self, int weekDay):
        return _sseImpl.isWeekEnd(weekDay)

    def __richcmp__(self, right, int op):
        if op == 2:
            return isinstance(right, ChinaCFFEXImpl)

cdef int EasterMonday[299]
EasterMonday[:] = [
    98, 90, 103, 95, 114, 106, 91, 111, 102,  # 1901-1909
    87, 107, 99, 83, 103, 95, 115, 99, 91, 111,  # 1910-1919
    96, 87, 107, 92, 112, 103, 95, 108, 100, 91,  # 1920-1929
    111, 96, 88, 107, 92, 112, 104, 88, 108, 100,  # 1930-1939
    85, 104, 96, 116, 101, 92, 112, 97, 89, 108,  # 1940-1949
    100, 85, 105, 96, 109, 101, 93, 112, 97, 89,  # 1950-1959
    109, 93, 113, 105, 90, 109, 101, 86, 106, 97,  # 1960-1969
    89, 102, 94, 113, 105, 90, 110, 101, 86, 106,  # 1970-1979
    98, 110, 102, 94, 114, 98, 90, 110, 95, 86,  # 1980-1989
    106, 91, 111, 102, 94, 107, 99, 90, 103, 95,  # 1990-1999
    115, 106, 91, 111, 103, 87, 107, 99, 84, 103,  # 2000-2009
    95, 115, 100, 91, 111, 96, 88, 107, 92, 112,  # 2010-2019
    104, 95, 108, 100, 92, 111, 96, 88, 108, 92,  # 2020-2029
    112, 104, 89, 108, 100, 85, 105, 96, 116, 101,  # 2030-2039
    93, 112, 97, 89, 109, 100, 85, 105, 97, 109,  # 2040-2049
    101, 93, 113, 97, 89, 109, 94, 113, 105, 90,  # 2050-2059
    110, 101, 86, 106, 98, 89, 102, 94, 114, 105,  # 2060-2069
    90, 110, 102, 86, 106, 98, 111, 102, 94, 114,  # 2070-2079
    99, 90, 110, 95, 87, 106, 91, 111, 103, 94,  # 2080-2089
    107, 99, 91, 103, 95, 115, 107, 91, 111, 103,  # 2090-2099
    88, 108, 100, 85, 105, 96, 109, 101, 93, 112,  # 2100-2109
    97, 89, 109, 93, 113, 105, 90, 109, 101, 86,  # 2110-2119
    106, 97, 89, 102, 94, 113, 105, 90, 110, 101,  # 2120-2129
    86, 106, 98, 110, 102, 94, 114, 98, 90, 110,  # 2130-2139
    95, 86, 106, 91, 111, 102, 94, 107, 99, 90,  # 2140-2149
    103, 95, 115, 106, 91, 111, 103, 87, 107, 99,  # 2150-2159
    84, 103, 95, 115, 100, 91, 111, 96, 88, 107,  # 2160-2169
    92, 112, 104, 95, 108, 100, 92, 111, 96, 88,  # 2170-2179
    108, 92, 112, 104, 89, 108, 100, 85, 105, 96,  # 2180-2189
    116, 101, 93, 112, 97, 89, 109, 100, 85, 105  # 2190-2199
]

cdef class WestenImpl(CalendarImpl):
    cdef bint isWeekEnd(self, int weekDay):
        return weekDay == Weekdays.Saturday or weekDay == Weekdays.Sunday

    cdef int easterMonday(self, int year):
        return EasterMonday[year - 1901]

cdef class TargetImpl(WestenImpl):
    def __init__(self):
        pass

    cdef bint isBizDay(self, Date date):
        cdef int w = date.weekday()
        cdef int d = date.dayOfMonth()
        cdef int dd = date.dayOfYear()
        cdef int m = date.month()
        cdef int y = date.year()
        cdef int em = self.easterMonday(y)

        if (self.isWeekEnd(w)
                or (d == 1 and m == Months.January)
                or (dd == em - 3 and y >= 2000)
                or (dd == em and y >= 2000)
                or (d == 1 and m == Months.May and y >= 2000)
                or (d == 25 and m == Months.December)
                or (d == 26 and m == Months.December and y >= 2000)
                or (d == 31 and m == Months.December and (y == 1998 or y == 1999 or y == 2001))):
            return False
        return True

    def __richcmp__(self, right, int op):
        if op == 2:
            return isinstance(right, TargetImpl)

cdef dict _holDict = {'china.sse': ChinaSseImpl,
                      'china.ib': ChinaIBImpl,
                      'china.cffex': ChinaCFFEXImpl,
                      'target': TargetImpl,
                      'null': NullCalendar,
                      'nullcalendar': NullCalendar}



# -*- coding: utf-8 -*-

from ultron.sentry.Math.Distributions.NormalDistribution cimport InverseCumulativeNormal as InvNormImpl


cdef class IAccumulator(object):
    pass


cdef class Accumulator(IAccumulator):

    cdef public bint _isFull
    cdef public list _dependency
    cdef public size_t _window

    cpdef bint isFull(self)
    cpdef push(self, dict data)
    cpdef double result(self)
    cpdef transform(self, data, str name=*, bint to_sort=*)

cdef bint isanumber(a)
cpdef build_holder(name)


cdef class Negative(Accumulator):

    cdef Accumulator _inner

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class CombinedValueHolder(Accumulator):

    cdef Accumulator _left
    cdef Accumulator _right

    cpdef push(self, dict data)


cdef class AddedValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class MinusedValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class MultipliedValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class DividedValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class LtOperatorValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class LeOperatorValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class GtOperatorValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class GeOperatorValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class EqOperatorValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class NeOperatorValueHolder(CombinedValueHolder):

    cpdef double result(self)


cdef class Identity(Accumulator):

    cdef double _value

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Current(Accumulator):

    cdef double _current

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Latest(Accumulator):

    cdef double _latest

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class CompoundedValueHolder(Accumulator):

    cdef Accumulator _left
    cdef Accumulator _right

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class IIF(Accumulator):

    cdef Accumulator _cond
    cdef Accumulator _left
    cdef Accumulator _right

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class BasicFunction(Accumulator):

    cdef double _origValue
    cdef Accumulator _inner

    cpdef push(self, dict data)


cdef class Exp(BasicFunction):

    cpdef double result(self)


cdef class Log(BasicFunction):

    cpdef double result(self)


cdef class Sqrt(BasicFunction):

    cpdef double result(self)


cdef class Pow(BasicFunction):

    cdef public double _n

    cpdef double result(self)


cdef class Abs(BasicFunction):

    cpdef double result(self)


cdef class Sign(BasicFunction):

    cpdef double result(self)


cdef class Acos(BasicFunction):

    cpdef double result(self)


cdef class Acosh(BasicFunction):

    cpdef double result(self)


cdef class Asin(BasicFunction):

    cpdef double result(self)


cdef class Asinh(BasicFunction):

    cpdef double result(self)


cdef class NormInv(BasicFunction):

    cdef InvNormImpl _inv

    cpdef double result(self)


cdef class Ceil(BasicFunction):

    cpdef double result(self)


cdef class Floor(BasicFunction):

    cpdef double result(self)


cdef class Round(BasicFunction):

    cpdef double result(self)
    
cdef class Sigmoid(BasicFunction):

    cpdef double result(self)
    
cdef class Tanh(BasicFunction):

    cpdef double result(self)


# -*- coding: utf-8 -*-

from enum import Enum
from enum import unique


class StrEnum(str, Enum):
    pass


@unique
class Factors(StrEnum):
    CLOSE = 'close'
    PE = 'pe'
    OPEN = 'open'
    VOLUME = 'volume'
    HIGH = 'high'
    LOW = 'low'



# -*- coding: utf-8 -*-
import json
import logging
import sys

import gevent
import time

from redis.client import pairs_to_dict
# from rediscluster import StrictRedisCluster
from redis import Redis, ConnectionPool, ResponseError, ConnectionError, TimeoutError

log = logging.getLogger(__name__)

class SingleRedisClient(Redis):

    def __init__(self, *args, **kwargs):
        self.status = type('RedisStatus', (), {'alive_timestamp': 0})
        startup_nodes = kwargs.get('startup_nodes')
        if startup_nodes:
            host = startup_nodes[0].get('host')
            password = startup_nodes[0].get('password')
            if 'redis://' not in host:
                kwargs['host'] = host
                kwargs['port'] = startup_nodes[0].get('port')
                if sys.version > '3':
                    kwargs.pop('startup_nodes')
                else:
                    del kwargs['startup_nodes']
                kwargs['password'] = password
            else:
                url = host.format(password=password)
                connection_pool = ConnectionPool.from_url(url, db=0, **kwargs)
                kwargs['connection_pool'] = connection_pool
                if sys.version > '3':
                    kwargs.pop('startup_nodes')
                else:
                    del kwargs['startup_nodes']
        else:
            host = kwargs.get('host')
            port = kwargs.get('port')
            password = kwargs.get('password')
            super(SingleRedisClient, self).__init__(
                host=host, port=port, password=password,
                max_connections=128, decode_responses=True
            )
        self.set_response_callback('GET', self._get)
        self.set_response_callback('HGETALL', self._hgetall)
        self.set_response_callback('HGET', self._hget)
        self.set_response_callback('HMGET', self._hmget)
        gevent.spawn(self._alive_check)
        gevent.sleep()

    # COMMAND EXECUTION AND PROTOCOL PARSING
    def execute_command(self, *args, **options):
        "Execute a command and return a parsed response"
        pool = self.connection_pool
        command_name = args[0]
        connection = pool.get_connection(command_name, **options)
        try:
            connection.send_command(*args)
            if command_name.upper() in ['HGETALL', 'HGET', 'HMGET', 'GET']:
                options['args'] = args
            return self.parse_response(connection, command_name, **options)
        except (ConnectionError, TimeoutError) as e:
            connection.disconnect()
            if not connection.retry_on_timeout and isinstance(e, TimeoutError):
                raise
            connection.send_command(*args)
            return self.parse_response(connection, command_name, **options)
        finally:
            pool.release(connection)

    @staticmethod
    def _get(response, **option):
        field = option.get('args')[1].split(':')[-1]
        if response is not None:
            field_prefix = field[:2]
            if field_prefix == 'i_':
                response = int(response)
            elif field_prefix == 'f_':
                response = float(response)
            elif field_prefix == 'b_':
                response = response not in ['0', 'False', 'false']
            elif field_prefix in ['d_', 'l_']:
                response = json.loads(response)
        return response

    @staticmethod
    def _hgetall(response, **option):
        response = pairs_to_dict(response)
        for k, v in response.items():
            if v is None:
                response[k] = v
                continue
            field_prefix = k[:2]
            if field_prefix == 'i_':
                response[k] = int(v)
            elif field_prefix == 'f_':
                response[k] = float(v)
            elif field_prefix == 'b_':
                response[k] = v not in ['0', 'False', 'false']
            elif field_prefix in ['d_', 'l_']:
                response[k] = json.loads(v)
        return response

    @staticmethod
    def _hget(response, **option):
        field = option.get('args')[2]
        if response is not None:
            field_prefix = field[:2]
            if field_prefix == 'i_':
                response = int(response)
            elif field_prefix == 'f_':
                response = float(response)
            elif field_prefix == 'b_':
                response = response not in ['0', 'False', 'false']
            elif field_prefix in ['d_', 'l_']:
                response = json.loads(response)
        return response

    @staticmethod
    def _hmget(response, **option):
        _response = {}
        fields = option.get('args')[2:]
        for field, value in zip(fields, response):
            if value is None:
                _response[field] = value
                continue
            field_prefix = field[:2]
            if field_prefix == 'i_':
                _response[field] = int(value)
            elif field_prefix == 'f_':
                _response[field] = float(value)
            elif field_prefix == 'b_':
                _response[field] = value not in ['0', 'False', 'false']
            elif field_prefix in ['d_', 'l_']:
                _response[field] = json.loads(value)
            else:
                _response[field] = value
        return _response

    def _alive_check(self):
        """
        Redis 存活检测
        """
        while True:
            try:
                if self.ping():
                    self.status.alive_timestamp = int(time.time())
            except Exception as e:
                log.exception(e)
                log.error('Redis Connection Exception.')
            gevent.sleep(3)

    def get_connection_status(self):
        return self.status

    def robust(self, func, *args, **kwargs):
        """
        对命令进行bobust的封装
        :param func:
        :param args:
        :param kwargs:
        :return:
        """
        try:
            return func(*args, **kwargs)
        except ResponseError as e:
            log.exception(e)
            log.debug(e)
            return None
        except Exception as e:
            log.exception(e)
            log.debug('Try Again.')
            if self.connection_pool._created_connections != 0:
                self.connection_pool.reset()
            gevent.sleep(2)
            return self.robust(func, *args, **kwargs)

    def setex(self, name, time, value):
        return super(SingleRedisClient, self).setex(name, value, time)

    def smadd(self, name, values):
        '''
        批量sadd
        '''
        with self.pipeline() as ppl:
            for value in values:
                ppl.sadd(name, value)
            ppl.execute()

    def smpop(self, name, count):
        '''
        批量spop
        '''
        with self.pipeline() as ppl:
            for _ in range(count):
                ppl.spop(name)
            ppl.execute()

    def hmdel(self, name, values):
        '''
        批量hdel
        '''
        with self.pipeline() as ppl:
            for value in values:
                ppl.hdel(name, value)
            ppl.execute()

    def hmove(self, old_name, new_name, key, value):
        with self.pipeline() as ppl:
            if old_name:
                ppl.hdel(old_name, key)
            ppl.hset(new_name, key, value)
            ppl.execute()

    def psubscribe(self, pattern):
        '''
        匹配订阅
        '''
        ps = self.pubsub()
        ps.psubscribe(pattern)
        log.info('Subscribe %s...' % pattern)
        for item in ps.listen():
            yield item
        ps.unsubscribe('spub')
        log.warning('Subscribe Was Exit.')

    def set_the_hash_value_for_the_hash(self, name, key, value_name, value_key, value):
        with self.pipeline() as ppl:
            ppl.hset(value_name, value_key, value)
            ppl.hset(name, key, value_name)
            ppl.execute()

    def get_the_hash_value_for_the_hash(self, name, key, value_key=None):
        value_name = self.hget(name, key)
        value = self.hget(value_name, value_key) if value_key else self.hgetall(value_name)
        return value

    def clean(self, pattern='*'):
        def _clean(_pattern):
            keys = self.keys(pattern)
            if not keys:
                return True
            self.delete(*keys)
            return True
        return self.robust(_clean, pattern)
    
    def hmgetall(self, *names):
        def _hmgetall(*_names):
            with self.pipeline(transaction=False) as ppl:
                for name in _names:
                    ppl.hgetall(name)
                ret = ppl.execute()
            return ret
        return self.robust(_hmgetall, *names)

    def hmget_all(self, names, field='all'):
        def _hmget_all(_names, _field):
            with self.pipeline(transaction=False) as ppl:
                for name in _names:
                    ppl.hget(name, _field)
                ret = ppl.execute()
            return ret
        return self.robust(_hmget_all, names, field)

    @staticmethod
    def timer(seconds, callback, *args, **kwargs):
        def _timer(_callback, *_args, **_kwargs):
            callback(*_args, **_kwargs)

        gevent.spawn_later(seconds, _timer, callback, *args, **kwargs)
        gevent.sleep()


RedisClient = SingleRedisClient



# -*- coding: utf-8 -*-

import abc
from typing import Union
import cvxpy as cp
import numpy as np

def require(cond: bool, exception=AssertionError, msg: str = None):
    if not cond:
        raise exception(msg)


def is_close(x: float, y: float, absolute: float = 1e-12) -> bool:
    return abs(x - y) <= absolute



class _IOptimizer(abc.ABC):
    def __init__(self,
                 cost: np.ndarray,
                 cons_matrix: np.ndarray = None,
                 lower_bound: Union[float, np.ndarray] = None,
                 upper_bound: Union[float, np.ndarray] = None):
        self._n = len(cost)
        self._cons_matrix = cons_matrix
        self._lower_bound = lower_bound
        self._upper_bound = upper_bound
        self._cost = cost
        
    def _prepare(self):
        x = cp.Variable(self._n)
        constraints = []
        if self._lower_bound is not None:
            require(isinstance(self._lower_bound, float) or len(self._lower_bound) == self._n,
                    ValueError,
                    "lower bounds must be a single value or an array with same size as x")
            constraints.append(x >= self._lower_bound)

        if self._upper_bound is not None:
            require(isinstance(self._upper_bound, float) or len(self._upper_bound) == self._n,
                    ValueError,
                    "upper bounds must be a single value or an array with same size as x")
            constraints.append(x <= self._upper_bound)

        if self._cons_matrix is not None:
            require(self._cons_matrix.shape[1] == self._n + 2,
                    ValueError,
                    "constraints must be a matrix with size as x + 2")
            constraints.append(self._cons_matrix[:, :self._n] @ x >= self._cons_matrix[:, self._n])
            constraints.append(self._cons_matrix[:, :self._n] @ x <= self._cons_matrix[:, self._n + 1])

        return x, constraints